/*********************************************************************
 *                                                                   *
 * Copyright (C) LANSA Group. All rights reserved                    *
 *                                                                   *
 *********************************************************************/

/**
 * @fileOverview Contains the core extensions required for default rendering and AutoGUI
 */

// Application base properties
function DEFINE_EXTENSION_axAppBaseProperties()
{
	var definition = {
		name: "axappx",
		display: TEXT.dev("axApplicationBasicProperties.display"),
		type: "internal",
		helpID: "axApplicationBasicProperties",
		allowCustomise: false,
		properties: {
			defaultTheme: {
				type: "String",
				description: TEXT.dev("axApplicationBasicProperties.defaultTheme")
			},
			autoGUIOn: {
				type: "Boolean",
				defaultStatic: true,
				description: TEXT.dev("axApplicationBasicProperties.autoGUIOn")
			},
			moveableWindows: {
				type: "Boolean",
				defaultStatic: true,
				description: TEXT.dev("axApplicationBasicProperties.moveableWindows")
			},
			strictLayoutGridForDbcs: {
				type: "Boolean",
				description: TEXT.dev("axApplicationBasicProperties.strictLayoutGridForDbcs")
			},
			onApplicationStart: {
				type: "Event",
				description: TEXT.dev("axApplicationBasicProperties.onApplicationStart")
			},
			onApplicationEnd: {
				type: "Event",
				description: TEXT.dev("axApplicationBasicProperties.onApplicationEnd")
			},
			onSignOn: {
				type: "Event",
				description: TEXT.dev("axApplicationBasicProperties.onSignOn")
			},
			onSignOff: {
				type: "Event",
				description: TEXT.dev("axApplicationBasicProperties.onSignOff")
			}
		},
		init: function(app)
		{
			if (!app._appStarted) 
			{
				this.raiseEvent("onApplicationStart");
				app._appStarted = true;
			}
			
			// Add listener for screen arrived events.
			AXES.Terminal.addListener("ScreenArrived", this._onScreenArrived, this);
			this._signedOn = false;
			
			var defaultTheme = this.getPropertyValue("defaultTheme");
			if (defaultTheme.length > 0) app.currentTheme(defaultTheme);
			
			this._prevAutoGUI = AXES.Layout.AutoGUI.autoGUIOn;
			AXES.Layout.AutoGUI.setAutoGUI(this.getPropertyValue("autoGUIOn"), false);
			
			AXES.Layout.moveableWindows = this.getPropertyValue("moveableWindows");
			this._appSettings = app;
		},
		destroy: function(app)
		{
			AXES.Terminal.removeListener("ScreenArrived", this._onScreenArrived);
			
			AXES.Layout.AutoGUI.setAutoGUI(this._prevAutoGUI, false);
			
			if (app.shuttingDown) 
			{
				if (app._signedOn) this._onSignOff();
				
				if (app._appStarted) 
				{
					this.raiseEvent("onApplicationEnd");
					app._appStarted = false;
				}
			}
			delete this._appSettings;
		},
		_onScreenArrived: function(eventObject)
		{
			if (AXES.Terminal.isLogin) 
			{
				// If it is the login screen and have been logged in, we have signed out.
				if (this._appSettings._signedOn) this._onSignOff();
			}
			else 
			{
				// If it is not the login screen and have not been logged in, we have signed in.	
				if (!this._appSettings._signedOn) this._onSignOn()
			}
		},
		_onSignOn: function()
		{
			this._appSettings._signedOn = true;
			this.raiseEvent("onSignOn");
		},
		_onSignOff: function()
		{
			this._appSettings._signedOn = false;
			this.raiseEvent("onSignOff");
		}
	};
	AXES.Extensions.add(definition);
}
DEFINE_EXTENSION_axAppBaseProperties();

// Application styles extension
function DEFINE_EXTENSION_axAppStyles()
{
    defineStyleClass();

    var extension =
    {
        name          : "axAppStyles",
        type          : "application",
        display       : TEXT.dev( "axAppStyles.display" ),
        helpID        : "axApplicationStyleCollection",
        allowCustomise: false,
        properties    :
        {
            interface:
            {
                type         : "String",
                defaultStatic: "list",
                description  : TEXT.dev( "axAppStyles.interface" ),
                items        : [
                    {
                        value: "list",
                        text : TEXT.dev( "axAppStyles.interface.list" )
                    },
                    {
                        value: "json",
                        text : TEXT.dev( "axAppStyles.interface.json" )
                    }
                ],
                noDynamic    : true
            },
			styles: {
				type: "axAppStyle",
				isList: true,
				immediateCommit: true,
				noDynamic: true,
				description: TEXT.dev("axAppStyles.styles")
			}
		},
		_initDone: false,
		init: function(app)
		{
			this._initDone = true;
			this.render(app);
		},
		render: function(app)
		{
			if (!this._initDone) return;
			this._removeCurrentSS();
			var ssPath = app.getStyleSheetPath();
			if (ssPath != "") this._styleSheet = AXES.Utils.LazyLoader.load(ssPath + "?_=" + (new Date()).getTime(), "css");
		},
		destroy: function(element, domElement)
		{
			this._removeCurrentSS();
		},
		_removeCurrentSS: function()
		{
			var ss = this._styleSheet;
			if (ss) 
			{
				if (ss.parentNode) ss.parentNode.removeChild(ss);
				delete this._styleSheet;
			}
		}
	}
	
	//
	function defineStyleClass()
	{
		var itemClass = function()
		{
		};
		itemClass.prototype.toString = function()
		{
			var name = this.getPropertyValue("name").trim();
			var theme = this.getPropertyValue("theme").trim();
			return theme.length > 0 ? theme + " - " + name : name;
		};
		
		//
		var styleForEnum = [{
			value: "manual",
			text: ""
		}, {
			value: "all",
			text: TEXT.dev("axAppStyles.styleFor.all")
		}, {
			value: "body",
			text: TEXT.dev("axAppStyles.styleFor.window")
		}, {
			value: "html",
			text: TEXT.dev("axAppStyles.styleFor.element")
		}];
		
		var for5250AttributesEnum = [{
			value: "20i",
			text: TEXT.dev("x20i - Input - Green")
		}, {
			value: "21i",
			text: TEXT.dev("x21i - Input - Green/Reverse Image")
		}, {
			value: "22i",
			text: TEXT.dev("x22i - Input - White")
		}, {
			value: "23i",
			text: TEXT.dev("x23i - Input - White/Reverse Image")
		}, {
			value: "24i",
			text: TEXT.dev("x24i - Input - Green/Underscore")
		}, {
			value: "25i",
			text: TEXT.dev("x25i - Input - Green/Underscore/Reverse Image")
		}, {
			value: "26i",
			text: TEXT.dev("x26i - Input - White/Underscore")
		}, {
			value: "27i",
			text: TEXT.dev("x27i - Input - Nondisplay")
		}, {
			value: "28i",
			text: TEXT.dev("x28i - Input - Red")
		}, {
			value: "29i",
			text: TEXT.dev("x29i - Input - Red/Reverse Image")
		}, {
			value: "2Ai",
			text: TEXT.dev("x2Ai - Input - Red/Blink")
		}, {
			value: "2Bi",
			text: TEXT.dev("x2Bi - Input - Red/Reverse Image/Blink")
		}, {
			value: "2Ci",
			text: TEXT.dev("x2Ci - Input - Red/Underscore")
		}, {
			value: "2Di",
			text: TEXT.dev("x2Di - Input - Red/Underscore/Reverse Image")
		}, {
			value: "2Ei",
			text: TEXT.dev("x2Ei - Input - Red/Underscore/Blink")
		}, {
			value: "2Fi",
			text: TEXT.dev("x2Fi - Input - Nondisplay")
		}, {
			value: "30i",
			text: TEXT.dev("x30i - Input - Turquoise/Column Separators")
		}, {
			value: "31i",
			text: TEXT.dev("x31i - Input - Turquoise/Column Separators/Reverse Image")
		}, {
			value: "32i",
			text: TEXT.dev("x32i - Input - Yellow/Column Separators")
		}, {
			value: "33i",
			text: TEXT.dev("x33i - Input - Yellow/Column Separators/Reverse Image")
		}, {
			value: "34i",
			text: TEXT.dev("x34i - Input - Turquoise/Underscore")
		}, {
			value: "35i",
			text: TEXT.dev("x35i - Input - Turquoise/Underscore/Reverse Image")
		}, {
			value: "36i",
			text: TEXT.dev("x36i - Input - Yellow/Underscore")
		}, {
			value: "37i",
			text: TEXT.dev("x37i - Input - Nondisplay")
		}, {
			value: "38i",
			text: TEXT.dev("x38i - Input - Pink")
		}, {
			value: "39i",
			text: TEXT.dev("x39i - Input - Pink/Reverse Image")
		}, {
			value: "3Ai",
			text: TEXT.dev("x3Ai - Input - Blue")
		}, {
			value: "3Bi",
			text: TEXT.dev("x3Bi - Input - Blue/Reverse Image")
		}, {
			value: "3Ci",
			text: TEXT.dev("x3Ci - Input - Pink/Underscore")
		}, {
			value: "3Di",
			text: TEXT.dev("x3Di - Input - Pink/Underscore/Reverse Image")
		}, {
			value: "3Ei",
			text: TEXT.dev("x3Ei - Input - Blue/Underscore")
		}, {
			value: "3Fi",
			text: TEXT.dev("x3Fi - Input - Nondisplay")
		}, {
			value: "20o",
			text: TEXT.dev("x20o - Output - Green")
		}, {
			value: "21o",
			text: TEXT.dev("x21o - Output - Green/Reverse Image")
		}, {
			value: "22o",
			text: TEXT.dev("x22o - Output - White")
		}, {
			value: "23o",
			text: TEXT.dev("x23o - Output - White/Reverse Image")
		}, {
			value: "24o",
			text: TEXT.dev("x24o - Output - Green/Underscore")
		}, {
			value: "25o",
			text: TEXT.dev("x25o - Output - Green/Underscore/Reverse Image")
		}, {
			value: "26o",
			text: TEXT.dev("x26o - Output - White/Underscore")
		}, {
			value: "27o",
			text: TEXT.dev("x27o - Output - Nondisplay")
		}, {
			value: "28o",
			text: TEXT.dev("x28o - Output - Red")
		}, {
			value: "29o",
			text: TEXT.dev("x29o - Output - Red/Reverse Image")
		}, {
			value: "2Ao",
			text: TEXT.dev("x2Ao - Output - Red/Blink")
		}, {
			value: "2Bo",
			text: TEXT.dev("x2Bo - Output - Red/Reverse Image/Blink")
		}, {
			value: "2Co",
			text: TEXT.dev("x2Co - Output - Red/Underscore")
		}, {
			value: "2Do",
			text: TEXT.dev("x2Do - Output - Red/Underscore/Reverse Image")
		}, {
			value: "2Eo",
			text: TEXT.dev("x2Eo - Output - Red/Underscore/Blink")
		}, {
			value: "2Fo",
			text: TEXT.dev("x2Fo - Output - Nondisplay")
		}, {
			value: "30o",
			text: TEXT.dev("x30o - Output - Turquoise/Column Separators")
		}, {
			value: "31o",
			text: TEXT.dev("x31o - Output - Turquoise/Column Separators/Reverse Image")
		}, {
			value: "32o",
			text: TEXT.dev("x32o - Output - Yellow/Column Separators")
		}, {
			value: "33o",
			text: TEXT.dev("x33o - Output - Yellow/Column Separators/Reverse Image")
		}, {
			value: "34o",
			text: TEXT.dev("x34o - Output - Turquoise/Underscore")
		}, {
			value: "35o",
			text: TEXT.dev("x35o - Output - Turquoise/Underscore/Reverse Image")
		}, {
			value: "36o",
			text: TEXT.dev("x36o - Output - Yellow/Underscore")
		}, {
			value: "37o",
			text: TEXT.dev("x37o - Output - Nondisplay")
		}, {
			value: "38o",
			text: TEXT.dev("x38o - Output - Pink")
		}, {
			value: "39o",
			text: TEXT.dev("x39o - Output - Pink/Reverse Image")
		}, {
			value: "3Ao",
			text: TEXT.dev("x3Ao - Output - Blue")
		}, {
			value: "3Bo",
			text: TEXT.dev("x3Bo - Output - Blue/Reverse Image")
		}, {
			value: "3Co",
			text: TEXT.dev("x3Co - Output - Pink/Underscore")
		}, {
			value: "3Do",
			text: TEXT.dev("x3Do - Output - Pink/Underscore/Reverse Image")
		}, {
			value: "3Eo",
			text: TEXT.dev("x3Eo - Output - Blue/Underscore")
		}, {
			value: "3Fo",
			text: TEXT.dev("x3Fo - Output - Nondisplay")
		}];
		
		var itemClassDef = {
			name: {
				type: "String",
				noDynamic: true
			},
			styleFor: {
				type: "String",
				items: styleForEnum,
				defaultStatic: "",
				noDynamic: true
			},
			htmlTag: {
				type: "String",
				noDynamic: true
			},
			style: {
				type: "Style",
				noDynamic: true
			},
			theme: {
				type: "String",
				noDynamic: true
			},
			for5250Attributes: {
				type: "String",
				isList: true,
				items: for5250AttributesEnum
			}
		};
		
		AXES.Extensions.definePropertyType("axAppStyle", {
			objectClass: itemClass,
			objectDefinition: itemClassDef
		});
	}
	
	AXES.Extensions.add(extension);
}
DEFINE_EXTENSION_axAppStyles();

// Application terminal properties
function DEFINE_EXTENSION_axAppTerminalProperties()
{
	var extension = {
		name: "axtrm",
		display: TEXT.dev("axApplicationTerminalProperties.display"),
		type: "application",
		helpID: "axApplicationTerminalProperties",
		allowCustomise: false,
		properties: {
			rowHeight: {
				type: "Integer",
				defaultStatic: 15,
				noDynamic: true,
				description: TEXT.dev("axApplicationTerminalProperties.rowHeight")
			},
			spaceBetweenRows: {
				type: "Integer",
				defaultStatic: 3,
				noDynamic: true,
				description: TEXT.dev("axApplicationTerminalProperties.spaceBetweenRows")
			},
			columnWidth: {
				type: "Integer",
				defaultStatic: 8,
				noDynamic: true,
				description: TEXT.dev("axApplicationTerminalProperties.columnWidth")
			},
			spaceBetweenColumns: {
				type: "Integer",
				defaultStatic: 0,
				noDynamic: true,
				description: TEXT.dev("axApplicationTerminalProperties.spaceBetweenColumns")
			}
		},
		init: function()
		{
			var l = AXES.Layout;
			this._prevSettings = [l.characterSize, l.spaceBetweenRows, l.spaceBetweenCols];
			
			l.characterSize = {
				width: this.getPropertyValue("columnWidth"),
				height: this.getPropertyValue("rowHeight"),
				locked: true
			}
			l.spaceBetweenRows = this.getPropertyValue("spaceBetweenRows");
			l.spaceBetweenCols = this.getPropertyValue("spaceBetweenColumns");
			
		},
		destroy: function()
		{
			var l = AXES.Layout;
			var p = this._prevSettings;
			
			if (p) 
			{
				l.characterSize = p[0];
				l.spaceBetweenRows = p[1];
				l.spaceBetweenCols = p[2];
				
				delete this._prevSettings;
			}
		}
	};
	
	AXES.Extensions.add(extension);
	//AXES.Application.TerminalSettingsExtensionDefinition = extension;
}
DEFINE_EXTENSION_axAppTerminalProperties();

// Screen basic properties
function DEFINE_EXTENSION_axScreenBaseProperties()
{
	var initFocusModeEnum = [{
		value: "default",
		text: TEXT.dev("axScreenBasicProperties.initFocusMode.5250")
	}, {
		value: "override",
		text: TEXT.dev("axScreenBasicProperties.initFocusMode.tabIndex")
	}];
	
	var extension = {
		name: "axsx",
		display: TEXT.dev("axScreenBasicProperties.display"),
		displayLong: TEXT.dev("axScreenBasicProperties.displayLong"),
		helpID: "axScreenBasicProperties",
		type: "internal",
		renderStage: "preRender",
		properties: {
			name: {
				type: "ScreenName",
				description: TEXT.dev("axScreenBasicProperties.name")
			},
			style: {
				type: "Style",
				description: TEXT.dev("axScreenBasicProperties.style")
			},
			firstElementToFocus: {
				type: "String",
				defaultStatic: "default",
				items: initFocusModeEnum,
				description: TEXT.dev("axScreenBasicProperties.initFocusMode")
			},
			useTerminalStyles: {
				type: "Boolean",
				defaultStatic: false,
				noDynamic: true,
				description: TEXT.dev("axScreenBasicProperties.useTerminalStyles")
			},
			disableTerminalKeys: {
				type: "Boolean",
				defaultStatic: false,
				noDynamic: true,
				description: TEXT.dev("axScreenBasicProperties.disableTerminalKeys")
			},
			limitedCursorSupport: {
				type: "Boolean",
				defaultStatic: false,
				noDynamic: true,
				description: TEXT.dev("axScreenBasicProperties.limitedCursorSupport")
			},
			disableAutoGUI: {
				type: "Boolean",
				defaultStatic: false,
				noDynamic: true,
				description: TEXT.dev("axScreenBasicProperties.disableAutoGUI")
			},
			onArrive: {
				type: "Event",
				description: TEXT.dev("axScreenBasicProperties.onArrive")
			},
			onLeave: {
				type: "Event",
				description: TEXT.dev("axScreenBasicProperties.onLeave")
			}
		},
		init: function(terminalContentDiv)
		{
			var $tc = $(terminalContentDiv);
			this.backupStyle(AXES.Layout.displayArea);
			
			if (!this.getPropertyValue("useTerminalStyles")) 
			{
				// Turn off terminal styles by removing the "axdefault" class. Styles that
				// can be turned off include it in their selector. Styles that are essential
				// to correct construction are not affected.
				AXES.Layout.displayArea.removeClass("axdefault");
				this._styleSheets = [];
			}
			
			// apply screen style
			this.applyStyle2(this.getPropertyValue("style"), AXES.Layout.displayArea);
			
			// screen events
			AXES.Terminal.addListener("ScreenArrived", this._onScreenArrive, this);
			AXES.Terminal.addListener("ScreenWillPost", this._onScreenLeave, this);
			
			if (this.getPropertyValue("disableAutoGUI")) 
			{
				this._prevAutoGUI = AXES.Layout.AutoGUI.autoGUIOn;
				AXES.Layout.AutoGUI.setAutoGUI(false, false);
			}
			
			//if (this.getPropertyValue("limitedCursorSupport")) 
			//{
				AXES.Layout.Cursor.init(this.getPropertyValue("limitedCursorSupport"));
				AXES.Layout.Cursor.show();
			//}
			// update main menu bar
			//this._hideAXESMainMenu(true);
		
			// TODO: Disable terminal keys property
			//screen._disableKeys = this.getPropertyValue("disableTerminalKeys");
		},
		
		_hideAXESMainMenu: function(hide)
		{
			var doc = AXES.getTsFrame().document;
			var items = ["menu_SelectTheme", "menu_CopyTable", "menu_CopyText", "menu_AutoGUI"];
			
			//
			if (hide) 
			{
				if (this._AXESMemoryMenuInfo == null) 
				{
					this._AXESMemoryMenuInfo = {};
					for (var i in items) 
					{
						var item = doc.getElementById(items[i]);
						if (item != null) 
						{
							this._AXESMemoryMenuInfo[item.id] = item.style.display;
							item.style.display = "none";
						}
					}
				}
			}
			else 
			{
				if (this._AXESMemoryMenuInfo != null) 
				{
					for (var id in this._AXESMemoryMenuInfo) 
					{
						var item = doc.getElementById(id);
						if (item != null) 
						{
							item.style.display = this._AXESMemoryMenuInfo[id];
						}
					}
					delete this._AXESMemoryMenuInfo;
				}
			}
		},
		destroy: function()
		{
			AXES.Layout.Cursor.hide();
			
			// Remove Terminal event listeners
			AXES.Terminal.removeListener("ScreenArrived", this._onScreenArrive);
			AXES.Terminal.removeListener("ScreenWillPost", this._onScreenLeave);
			
			// update main menu bar
			//this._hideAXESMainMenu(false);
			
			if (this._prevAutoGUI != undefined) 
			{
				AXES.Layout.AutoGUI.autoGUIOn = this._prevAutoGUI;
				delete this._prevAutoGUI;
			}
			
			// Restore the backed up style (except any height and width values)
			var s = AXES.Layout.displayArea[0].style;
			var h = s.height;
			var w = s.width;
			this.applyStyle2("", AXES.Layout.displayArea, true);
			s.height = h;
			s.width = w;
			
		},
		
		_onScreenArrive: function()
		{
			// TODO: firstElementToFocus property
			this.raiseEvent("onArrive");
		},
		
		_onScreenLeave: function(eventInfo)
		{
			var keyName = AXES.Keyboard.aidKeysByCode[eventInfo.aidKey];
			if (keyName == undefined) keyName = eventInfo.aidKey;
			
			var result = this.raiseEvent("onLeave", {
				key: keyName.toUpperCase()
			});
			
			// only cancel leave if return value is explicitly set to false
			result = (result !== false);
			return result;
		}
	};
	AXES.Extensions.add(extension);
}
DEFINE_EXTENSION_axScreenBaseProperties();

// Popup window properties
function DEFINE_EXTENSION_axPopupWindow()
{
	//
	var extension = {
		name: "axpopupwindow",
		display: TEXT.dev("axPopupWindow.display"),
		displayLong: TEXT.dev("axPopupWindow.displayLong"),
		helpID: "axPopupWindow",
		type: "screen",
		renderStage: "windowStart",
		properties: {
			style: {
				type: "Style",
				description: TEXT.dev("axPopupWindow.style")
			},
			backgroundStyle: {
				type: "Style",
				description: TEXT.dev("axPopupWindow.backgroundStyle")
			},
			windowTop: {
				type: "Integer",
				defaultStatic: -1,
				description: TEXT.dev("axPopupWindow.windowTop")
			},
			windowLeft: {
				type: "Integer",
				defaultStatic: -1,
				description: TEXT.dev("axPopupWindow.windowLeft")
			},
			windowWidth: {
				type: "Integer",
				defaultStatic: -1,
				description: TEXT.dev("axPopupWindow.windowWidth")
			},
			windowHeight: {
				type: "Integer",
				defaultStatic: -1,
				description: TEXT.dev("axPopupWindow.windowHeight")
			}
		},
		init: function(windowDiv)
		{
			if (windowDiv.id == "terminalContent") {
				AXES.SysTrace.warn(TEXT.Dev("Attempt to use the axPopupWindow eXtension on a screen with no window."));
				return;
			}
			this._windowDiv = windowDiv;
			this._backgroundDiv = $(windowDiv).children(".ax-terminal-window-background")[0];
			this.backupStyle(windowDiv);
			this.backupStyle(this._backgroundDiv);
		},
		
		render: function(windowDiv) 
		{
			if (!this._windowDiv) return;
			var l = this.getPropertyValue("windowLeft"),
			t = this.getPropertyValue("windowTop"),
			w = this.getPropertyValue("windowWidth"),
			h = this.getPropertyValue("windowHeight"),
			sWin = this.getPropertyValue("style"),
			sBgnd = this.getPropertyValue("backgroundStyle"),
			toDelete = ["left","top","width","height"],
			i, diff;
			
			for (i=0; i<3; i++) 
			{
				delete sWin[toDelete[i]];
				delete sBgnd[toDelete[i]];
			}
			
			if (l>=0) sWin.left = l + "px";
			if (t>=0) sWin.top = t + "px";
			
			if (w >= 0) 
			{
				diff = $(this._windowDiv).innerWidth() -  $(this._backgroundDiv).innerWidth();
				sBgnd.width = w + "px";
				sWin.width = (w + diff) + "px";
			}
			
			if (h >= 0) 
			{
				diff = $(this._windowDiv).innerHeight() -  $(this._backgroundDiv).innerHeight();
				sBgnd.height = h + "px";
				sWin.height = (h + diff) + "px";
			}
			
			this.applyStyle2(sBgnd, this._backgroundDiv);
			this.applyStyle2(sWin, this._windowDiv);
		},
		
		destroy: function()
		{
			if (!this._windowDiv) return;
			// Restore the backed up style
			this.applyStyle2("", this._windowDiv, true);
			this.applyStyle2("", this._backgroundDiv, true);
		}
	};
	AXES.Extensions.add(extension);
}
DEFINE_EXTENSION_axPopupWindow();

// Screen element (existing) basic properties
function DEFINE_EXTENSION_axScreenElBaseProperties()
{
	var definition = {
		name: "axex",
		display: TEXT.dev("axElementBasicProperties.display"),
		displayLong: TEXT.dev("axElementBasicProperties.displayLong"),
		type: "internal",
		helpID: "axElementBasicProperties",
		properties: {
			style: {
				type: "Style",
				noDynamic: true,
				description: TEXT.dev("axElementBasicProperties.style")
			},
			tabIndex: {
				type: "Integer",
				defaultStatic: 0,
				noDynamic: true,
				description: TEXT.dev("axElementBasicProperties.tabIndex")
			},
			visible: {
				type: "Boolean",
				defaultStatic: true,
				description: TEXT.dev("axElementBasicProperties.visible")
			},
			enabled: {
				type: "Boolean",
				defaultStatic: true,
				description: TEXT.dev("axElementBasicProperties.enabled")
			},
			tooltip: {
				type: "String",
				description: TEXT.dev("axElementBasicProperties.tooltip")
			},
			applyToInputOutput: {
				type: "Boolean",
				defaultStatic: false,
				noDynamic: true,
				description: TEXT.dev("axElementBasicProperties.applyInputOutput")
			}
		},
		init: function(field, fieldDiv)
		{
			this._field = field;
			this._fieldDiv = fieldDiv;
			this.backupStyle(fieldDiv);
		},
		render: function()
		{
			var style = this.getPropertyValue("style");
			if (!this.isDesignMode()) 
			{
				style.visibility = (this.getPropertyValue("visible") == false) ? "hidden" : "visible";
			}
			this.fixTS1FakeWindowPositions(style);
			this.applyStyle2(style, this._fieldDiv, true);
			
			var enabled = this.getPropertyValue("enabled");
			this._fieldDiv.disabled = (enabled == false);
			
			this._fieldDiv.title = this.getPropertyValue("tooltip");
		},
		destroy: function(element, htmlElement)
		{
			this._fieldDiv.title = "";
			delete this._field;
			delete this._fieldDiv;
		}
	};
	AXES.Extensions.add(definition);
}
DEFINE_EXTENSION_axScreenElBaseProperties();

// Screen element (new) basic properties
function DEFINE_EXTENSION_axCustScreenElBaseProperties()
{
	var definition = {
		name: "axuex",
		display: TEXT.dev("axUserElementBasicProperties.display"),
		displayLong: TEXT.dev("axUserElementBasicProperties.displayLong"),
		helpID: "axUserElementBasicProperties",
		type: "internal",
		properties: {
			name: {
				type: "String",
				noDynamic: true,
				description: TEXT.dev("axUserElementBasicProperties.name")
			},
			style: {
				type: "Style",
				noDynamic: true,
				description: TEXT.dev("axUserElementBasicProperties.style")
			},
			type: {
				type: "Integer",
				items: [{
					value: 1,
					text: TEXT.dev("axUserElementBasicProperties.type.input")
				}, {
					value: 2,
					text: TEXT.dev("axUserElementBasicProperties.type.output")
				}],
				defaultStatic: 1,
				noDynamic: true,
				description: TEXT.dev("axUserElementBasicProperties.type")
			},
			tabIndex: {
				type: "Integer",
				defaultStatic: 0,
				noDynamic: true,
				description: TEXT.dev("axUserElementBasicProperties.tabIndex")
			},
			visible: {
				type: "Boolean",
				defaultStatic: true,
				description: TEXT.dev("axUserElementBasicProperties.visible")
			},
			enabled: {
				type: "Boolean",
				defaultStatic: true,
				description: TEXT.dev("axUserElementBasicProperties.enabled")
			},
			tooltip: {
				type: "String",
				description: TEXT.dev("axUserElementBasicProperties.tooltip")
			}
		},
		init: function(field, fieldDiv)
		{
			this._field = field;
			this._fieldDiv = fieldDiv;
			this.backupStyle(fieldDiv);
			//var exts = element.getExtensions();
			//var drawBorder = exts.length == 1 && exts[0] == element.getBasePropExtension();
			//if (drawBorder) htmlElement.style.backgroundColor = "rgb(240,240,240)";
		},
		render: function()
		{
			this._field.symbolicName = this.getPropertyValue("name");
			
			var style = this.getPropertyValue("style");
			if (!this.isDesignMode()) 
			{
				style.visibility = (this.getPropertyValue("visible") == false) ? "hidden" : "";
			}
			this.fixTS1FakeWindowPositions(style);
			this.applyStyle2(style, this._fieldDiv, true);
			
			var enabled = this.getPropertyValue("enabled");
			this._fieldDiv.disabled = (enabled == false);
			
			this._fieldDiv.title = this.getPropertyValue("tooltip");
			
			switch (this.getPropertyValue("type")) {
				case 1:
					this._field.type = AXES.Terminal.Field.Types.INPUT;
					this._field.isInput = true;
					break;
				case 2:
					this._field.type = AXES.Terminal.Field.Types.OUTPUT;
					this._field.isInput = false;
					break;
			}
		},
		destroy: function()
		{
			this._fieldDiv.title = "";
			delete this._field;
			delete this._fieldDiv;
		}
	};
	AXES.Extensions.add(definition);
}
DEFINE_EXTENSION_axCustScreenElBaseProperties();

function DEFINE_EXTENSION_axDefault()
{
	var imeModeEnum = [{
		value: "default",
		text: TEXT.dev("axDefaultVisualization.imeMode.default")
	}, {
		value: "active",
		text: TEXT.dev("axDefaultVisualization.imeMode.active")
	}, {
		value: "inactive",
		text: TEXT.dev("axDefaultVisualization.imeMode.inactive")
	}];

	var definition = {
		/**
		 * Programmatic class name of this extension. It's not visible to end user
		 **/
		name: "axdv",
		group: "vis",
		/**
		 * Type of object this extension is applicable. Can be either "element" or "screen"
		 * If this extension applies to both element & screen, we put both in an array like so:
		 * type: ["element", "screen"]
		 **/
		type: "element",
		
		/**
		 * End-user visible name of this extension
		 **/
		display: TEXT.dev("axDefaultVisualization.display"),
		helpID: "axDefaultVisualization",
		
		properties: {
			value: {
				type: "String",
				description: TEXT.dev("axDefaultVisualization.value")
			},
			style: {
				type: "Style",
				description: TEXT.dev("axDefaultVisualization.style")
			},
			maxInputLength: {
				type: "Integer",
				defaultStatic: 0,
				description: TEXT.dev("axDefaultVisualization.maxInputLength")
			},
			selectContentOnFocus: {
				type: "Boolean",
				defaultStatic: true,
				description: TEXT.dev("axDefaultVisualization.selectContentOnFocus")
			},
			imeMode: {
				type: "String",
				defaultStatic: "default",
				description: TEXT.dev("axDefaultVisualization.imeMode"),
				items: imeModeEnum
			},
			removeCUADots: {
				type: "Boolean",
				description: TEXT.dev("axDefaultVisualization.removeCUADots")
			},
			AIDKeyOnDoubleClick: {
				type: "String",
				defaultStatic: "Enter",
				description: TEXT.dev("axDefaultVisualization.AIDKeyOnDoubleClick")
			}
		},
		
		init: function(field, fieldDiv)
		{
			this._field = field;
			this._fieldDiv = fieldDiv;
			this._content = null;
			this._addedToDOM = false;
			
			var newEl = null;
			var type = (field) ? field.type : AXES.Terminal.Field.Types.INPUT;
			switch (type) {
				case AXES.Terminal.Field.Types.OUTPUT:
					this._createTextField();
					break;
				case AXES.Terminal.Field.Types.INPUT:
					this._createInputField();
					break;
				case AXES.Terminal.Field.Types.SELECT:
					this._createSelectField();
					break;
				case AXES.Terminal.Field.Types.SCROLLBAR:
					this._createScrollbarField();
					break;
				case AXES.Terminal.Field.Types.HTML:
					this._createHtmlField();
					break;
				default:
			}
			
			if (this._content && (this._content.length > 0)) 
			{
				this._fieldDiv.appendChild(this._content[0]);
			}
			
			field.addListener("AddedToDOM", this._fieldAddedToDOM, this);
			
			this.backupStyle(this._content);
		},
		render: function(field, fieldDiv)
		{
			var type = (field) ? field.type : AXES.Terminal.Field.Types.INPUT;
			
			var newStyle = this.getPropertyValue("style");
			
			if (field.type == AXES.Terminal.Field.Types.INPUT) 
			{
				var maxLength = this.getPropertyValue("maxInputLength");
				if (typeof maxLength == "number" && !isNaN(maxLength) && maxLength > 0) 
				{
					this._content[0].setAttribute("maxlength", maxLength);
				} else {
					if (AXES.isTS2Engine && this.isDesignMode()) {
						var size = field.getSize();
						// set the size of the element to occupy the whole fieldDiv container
						newStyle = {
						        width: size.width,
						        height: size.height - 1
						};
					}
				}
				
				var imeMode = this.getPropertyValue("imeMode");
				if (imeMode != "default") newStyle["ime-mode"] = imeMode;
			}
			
			this._setValue();
			
			// Apply new styles
			this.applyStyle2(newStyle, this._content[0], true);
		},
		destroy: function()
		{
			if (this._content) this._content.remove();
			this._field.removeAllListenersByContext(this);
			delete this._content;
			delete this._field;
			delete this._fieldDiv;
		},
		_setValue: function()
		{
			var field = this._field, signedNum, val = this.getPropertyValue("value");
			if (!(AXES.Utils.isString(val) && val.length > 0)) {
				val = this._field.getValue();
				switch ((field) ? field.type : AXES.Terminal.Field.Types.INPUT) {
					case AXES.Terminal.Field.Types.OUTPUT:
						if (this.getPropertyValue("removeCUADots")) 
						{
							val = val.replace(/\s?(\. )*\.?$/, "");
						}
						break;
					case AXES.Terminal.Field.Types.INPUT:
						break;
						
					default:
				}				
			}
			
			switch ((field) ? field.type : AXES.Terminal.Field.Types.INPUT) {
				case AXES.Terminal.Field.Types.OUTPUT:
					this._content.text(val);
					break;
				case AXES.Terminal.Field.Types.INPUT:
					if (!this._content[0]._noFieldValueChangedEvent) {
						signedNum = (((field.fieldFormatWord & 0x700) >> 8) == 7);
						if (field.rightAlign || signedNum) {
						    this._masterVal = (field.rightAlign || signedNum) ? val.ltrim() : val.rtrim();
						    if (signedNum && (" -".indexOf(this._masterVal.slice(-1)) < 0)) {
						        this._masterVal = this._masterVal.replace(/ /g,'');
						        this._masterVal += " ";
						    }
						    this._content.val(this._masterVal);
						} else {
						    this._masterVal = val;
						    this._content.val(this._masterVal);
						}
					} else delete this._content[0]._noFieldValueChangedEvent;
					break;
				default:
			}

		},
		_createTextField: function()
		{
			var field = this._field;
			var newEl = document.createElement("span");
			newEl.className = "ax-terminal-textField";
			if (field instanceof AXES.Terminal.Field.TextField) 
			{
				newEl.className += " ax_x" + field.attribute.substring(1).toUpperCase() + "o";
			}
			
			this._content = $(newEl).bind("click", this, function(e)
			{
				e.data._field.setCursor();
			}).bind("dblclick", this, this._handleOutputDblClick);
			this._field.addListener("ValueChanged", this._fieldChangedText, this);
		},
		_createInputField: function()
		{
			var field = this._field, attr = parseInt("0" + field.attribute),
			isHidden = (attr == 0x27) || (attr == 0x2f) || (attr == 0x37) || (attr == 0x3f),
			newEl, css;
			if (field && field.height > 1 && (!isHidden)) 
			{
				newEl = document.createElement("textarea");
				newEl.setAttribute("cols", field.width);
				newEl.setAttribute("rows", field.height);
				newEl.style.height = field.height + "em";
			}
			else 
			{
				newEl = document.createElement("input");
				newEl.setAttribute("type", (AXES.isMobile && field.isNumeric) ? "text" : (isHidden ? "password" : "text"));
				if (field.len) newEl.setAttribute("maxlength", field.len);
				if (field.width) newEl.setAttribute("size", field.width);
			}

			newEl.className = "ax-terminal-inputField";
			if (field instanceof AXES.Terminal.Field.InputField) 
			{
				newEl.className += " ax_x" + field.attribute.substring(1).toUpperCase() + "i";
			}
			
			this._content = $(newEl);
			if (field) 
			{
				if (field.isFeatureInput) newEl.setAttribute("readonly", "readonly");
				css = { "ime-mode": field.imeMode };
				if (field.rightAlign) {
					css["text-align"] = "right";
					css["padding-right"] = "1px";
				}
				if (field.width) {
					css["width"] = AXES.Layout.characterDimensionsToPixels({width: field.width}, false, true).width;
					if (field.rightAlign) css["width"] -= 1;
					css["width"] += "px";
				}
				
				this._content.css(css).bind("input", this, this._handleInput)
				.bind("change", this, this._handleInput)
				.bind("focus", this, this._handleFocus)
				.bind("keydown", this, this._handleKeyDown)
				.bind("keypress", this, this._handleKeyPress)
				.bind("keyup", this, this._handleKeyUp);

				this.applyTabIndex(newEl, true);
				field.addListener("SetFocus", this.setFocus, this);
				field.addListener("ValueChanged", this._fieldChangedInput, this);
			}
		},

		_createSelectField: function()
		{
			var field = this._field;
			var cH = AXES.Layout.characterSize.height;
			var cW = AXES.Layout.characterSize.width;
			var xOffset = 0;
			var yOffset = 0;
			
			var wrapper = document.createElement("div");
			wrapper.setAttribute("id", field.id);
			var selectClass = (field.selectType == "Menu") ? "ax-terminal-selectMenu" : "ax-terminal-select";
			wrapper.className = selectClass;
			var wrSize = {
				width: field.width,
				height: field.height
			};
			wrSize = AXES.Layout.characterDimensionsToPixels(wrSize, true, true);
			wrapper.style.width = wrSize.width;
			wrapper.style.height = wrSize.height;
			
			if (field.scrollbar) 
			{
				var sb = this._drawScrollbar();
				$(sb).bind("click", this, this._handleScroll);
				wrapper.appendChild(sb);
			}
			
			for (var i = 0, opt; opt = field.selectItems[i]; i++) 
			{
				var selItem = document.createElement("span");
				selItem.setAttribute("id", field.id + "-" + opt.sequence);
				selItem.className = "ax-terminal-selectItem";
				selItem.style.top = ((opt.row - yOffset) * cH) + "px";
				selItem.style.left = ((opt.col - xOffset) * cW) + "px";
				
				switch (field.ind) {
					case "fld":
						var inpId = field.id + "-" + opt.sequence + "_input";
						var inp = document.createElement("input");
						var lbl = document.createElement("label");
						inp.setAttribute("id", inpId);
						inp.setAttribute("name", field.id);
						lbl.setAttribute("for", inpId);
						lbl.innerHTML = opt.text.rtrim();
						inp.type = (field.multi) ? "checkbox" : "radio";
						if (opt.selected) {
							inp.setAttribute("checked", "checked");
						}
						if (opt.byp) {
							inp.setAttribute("disabled", "disabled");
							lbl.className = "disabled";
						}
						inp.value = opt.sequence;
						selItem.appendChild(inp);
						selItem.appendChild(lbl);
						this.applyTabIndex(inp, true);
						break;
					case "btn":
						var inp = document.createElement("button");
						inp.value = opt.sequence;
						var btnText = opt.text.trim();
						if (btnText == "") {
							btnText = "&nbsp;";
						}
						inp.innerHTML = btnText;
						if (opt.selected) {
							inp.setAttribute("checked", "checked");
						}
						if (opt.byp) {
							inp.setAttribute("disabled", "disabled");
						}
						inp.style.width = (field.childWidth * cW) + "px";
						
						selItem.appendChild(inp);
						this.applyTabIndex(inp, true);
						break;
					default:
						var span = document.createElement("span");
						span.className = "ax-terminal-selectItem-text" + ((opt.selected) ? " ui-state-highlight" : "");
						span.innerHTML = opt.text.rtrim();
						selItem.appendChild(span);
						this.applyTabIndex(span, true);
						$(selItem).bind("dblclick", this, this._handleSelectDblClick);
				}
				
				$(selItem).data("ItemIndex", i).bind("click", this, this._handleSelectClick).bind("keydown", this, this._handleSelectKeyDown);
				wrapper.appendChild(selItem);
			}
			
			this._content = $(wrapper);
			field.addListener("SetFocus", this.setFocus, this);
			field.addListener("ValueChanged", this._fieldChangedSelect, this);
		},
		_createScrollbarField: function()
		{
			var sb = this._drawScrollbar();
			this._content = $(sb).bind("click", this, this._handleScroll);
		},
		_drawScrollbar: function()
		{
			var field = this._field;
			var cH = AXES.Layout.characterSize.height;
			var cW = AXES.Layout.characterSize.width;
			
			// Direction 1 = Up/Left
			// Direction 2 = Down/Right
			var sb = document.createElement("div");
			sb.className = "ax-terminal-scrollbar " + (field.scrollVertical ? "ax-terminal-scrollbar-vert" : "ax-terminal-scrollbar-horiz");

			var sbSize = { width: field.width, height: field.height };
			sbSize = AXES.Layout.characterDimensionsToPixels(sbSize, true, true);
			sb.style.width = field.scrollVertical ? "16px" : sbSize.width;
			sb.style.height = field.scrollVertical ? sbSize.height : "16px";
			
			var btn = document.createElement("div");
			btn.setAttribute("id", field.id + "scroll_pageDir1");
			btn.className = "ui-state-default ui-icon ax-terminal-scrollPageDir1";
			sb.appendChild(btn);
			btn = document.createElement("div");
			btn.setAttribute("id", field.id + "scroll_dir1");
			btn.className = "ui-state-default ui-icon ax-terminal-scrollDir1";
			sb.appendChild(btn);
			
			btn = document.createElement("div");
			btn.setAttribute("id", field.id + "scroll_dir2");
			btn.className = "ui-state-default ui-icon ax-terminal-scrollDir2";
			sb.appendChild(btn);
			btn = document.createElement("div");
			btn.setAttribute("id", field.id + "scroll_pageDir2");
			btn.className = "ui-state-default ui-icon ax-terminal-scrollPageDir2";
			sb.appendChild(btn);
			
			var thumb = document.createElement("div");
			thumb.className = "ax-terminal-scrollThumb";
			
			var bar = document.createElement("div");
			bar.className = "ax-terminal-scrollLine";
			thumb.appendChild(bar);
			var tick = document.createElement("div");
			tick.className = "ax-terminal-scrollTick";
			
			var tickPos = field.scrollPosition / field.scrollRows * 100;
			var tickSize = field.height / field.scrollRows * 100;
			if (field.scrollVertical) 
			{
				tick.style.top = tickPos + "%";
				tick.style.height = tickSize + "%";
			}
			else 
			{
				tick.style.left = tickPos + "%";
				tick.style.width = tickSize + "%";
			}
			thumb.appendChild(tick);
			
			sb.appendChild(thumb);
			return sb;
		},
		_createHtmlField: function()
		{
			var field = this._field, span = document.createElement("span");
			span.className = "ax-terminal-htmlField";
			try 
			{
				span.innerHTML = field.html;
			} catch (e) 
			{
				AXES.SysTrace.warn(TEXT.Dev("Invalid HTML supplied in DDS cannot be placed into the screen."));
				return;
			}
			this._content = $(span);
			
			if (field && field.scripts.length > 0) 
			{
				field.addListener("AddedToDOM", function(event)
				{
					for (var j = 0, code; code = event.field.scripts[j]; j++) 
					{
						try 
						{
							axExecuteScript("(function() { " + code + " })();");
						} catch (e) 
						{
							AXES.SysTrace.error(TEXT.Core("Error executing JavaScript embedded in DDS: %s1", e.message));
						}
					}
				}, this);
			}
		},
		/* Field events. These are events that are generated by the AXES.Terminal.Field
		 * object.  The Axes event mechanism allows the context of the event
		 * handler to be defined when it is registered.  So, unlike browser events,
		 * the "this" keyword in these handlers refers to this extension object.
		 */
		setFocus: function(event)
		{
			// May not be able to set the focus if the field is invisible
			try 
			{
				switch (event.field.type) {
					case AXES.Terminal.Field.Types.INPUT:
						if (event.offset != undefined) {
							this.offsetOnFocus = event.offset;
						}
						this._content[0].focus();
						break;
					case AXES.Terminal.Field.Types.SELECT:
						var el = this._content.find("span.ax-terminal-selectItem").find("input,button").get(event.offset ? event.offset : 0);
						el.focus();
						break;
					default:
				}
			}
			catch(e) {}
		},
		_fieldChangedText: function(event)
		{
			this._setValue();
		},
		_fieldChangedInput: function(event)
		{
			var fld = event.field;
			fld.mdt = true;
			this._setValue();
			var pling = $("span.ax-default-error-indicator", this._fieldDiv);
			if (fld.errorMessage != "") 
			{
				if (pling.length == 0) 
				{
					var inner = $(document.createElement("span")).addClass("ui-icon ui-icon-alert");
					pling = $(document.createElement("span")).addClass("ax-default-error-indicator ui-state-error").css({
						"display": "inline-block",
						"vertical-align": "top",
						"border": "none"
					}).append(inner).insertAfter(this._content);
				}
				pling.attr("title", fld.errorMessage);
			}
			else 
			{
				pling.remove();
			}
		},
		_fieldChangedSelect: function(event)
		{
			var field = event.field;
			switch (field.ind) {
				case "fld":
					this._content.find("ax-terminal-selectItem").each(function(){
						var $this = $(this), $inp = $this.find("input:checkbox,input:radio");
						$inp.attr("checked", (field.isItemSelected($this.data("ItemIndex"))) ? "checked" : "");
					});
					break;
				case "btn":
					break;
				default:
					this._content.find("span.ax-terminal-selectItem").each(function()
					{
						var $this = $(this), $span = $this.find("span.ax-terminal-selectItem-text");
						
						if (field.isItemSelected($this.data("ItemIndex"))) 
						{
							$span.addClass("ui-state-highlight");
						}
						else 
						{
							$span.removeClass("ui-state-highlight");
						}
					});
			}
		},
		_fieldAddedToDOM: function(event)
		{
			var $c = this._content;
			var newStyle = {};
			var size = this._field.getSize();
			// Extension elements size themselves. Only force the issue if the 
			// designer has explicitly set a size.
			var s = this._fieldDiv.style;
			if (s.width != "") 
			{
				$c.width(size.width - ($c.outerWidth() - $c.innerWidth()));
			}
			if (s.height != "") 
			{
				if (!$c.is("input,select")) 
				{
					$c.height(size.height - ($c.outerHeight() - $c.innerHeight()));
				}
			}
			//this.applyStyle2(newStyle, this._content, false);
		},
		/* Browser events on the field element.
		 * In these event handlers "this" refers to the browser element that
		 * received the event. Because of the way this extension is defined
		 * within an object, we cannot use closures to get access to the
		 * extension object.  To solve this the jQuery "bind" method is used
		 * to assign the event handlers. This method allows a "data" property to
		 * added to the event object. The "data" property is set to the extension.
		 */
		_handleSelectClick: function(e)
		{
			var field = e.data._field, index = $(this).data("ItemIndex");
			if (field.selectItems[index].byp) return;
			field.setCursor(index);
			switch (field.ind) {
				case "fld":
					if (field.multi) {
						field.toggleItemSelected(index);
					}
					else {
						field.setItemSelected(index, true);
					}
					break;
				default:
					field.setItemSelected(index, true);
			}
			if (field.autoPost) 
			{
				var aidKey = field.selectItems[index].aidKey;
				AXES.Terminal.doPost((aidKey) ? aidKey : AXES.Keyboard.aidKeys["PostData"]);
			}
		},
		_handleSelectDblClick: function(e)
		{
			var field = e.data._field, index = $(this).data("ItemIndex");
			if (field.selectItems[index].byp || field.autoPost) return;
			field.setCursor(index);
			field.setItemSelected(index, true);
			var aidKey = field.selectItems[index].aidKey;
			AXES.Terminal.doPost((aidKey) ? aidKey : AXES.Keyboard.aidKeys["PostData"]);
		},
		_handleSelectKeyDown: function(e)
		{
			var field = e.data._field;
			var index = $(this).data("ItemIndex");
			if (field.selectItems[index].byp) return;
			if (e.axKeyInfo && e.axKeyInfo.mappedTo) 
			{
				switch (e.axKeyInfo.mappedTo) {
					case "FieldExit":
					case "NextField":
						if (++index < field.selectItems.length) {
							field.setFocus(index);
							e.preventDefault();
							e.stopPropagation();
							e.axKeyInfo.preventKeypress = true;
						}
						break;
					case "PrevField":
						if (index > 0) {
							field.setFocus(index - 1);
							e.preventDefault();
							e.stopPropagation();
							e.axKeyInfo.preventKeypress = true;
						}
						break;
				}
			}
		},
		_handleScroll: function(e)
		{
			var field = e.data._field, t = e.target, tmp, action, key, o, s;
			field.setCursor();
			if ($(t).hasClass("ax-terminal-scrollLine")) {
				tmp = $(t).parent().children(".ax-terminal-scrollTick");
				o = tmp.offset();
				s = tmp.height();
				if (field.scrollVertical) {
					s = tmp.height();
					if (e.pageY < o.top) action = "pageDir1";
					if (e.pageY > o.top + s) action = "pageDir2";
				} else {
					s = tmp.width();
					if (e.pageX < o.left) action = "pageDir1";
					if (e.pageX > o.left + s) action = "pageDir2";
				}
			} else {
				tmp = t.id.split("_");
				action = tmp[tmp.length - 1];
			}
			switch (action) {
				case "pageDir1":
					if (field.scrollPosition <= 0) return;
					field.scrollAmount = 0;
					key = field.scrollVertical ? "RollDown" : "RollRight";
					break;
				case "dir1":
					if (field.scrollPosition <= 0) return;
					field.scrollAmount = 1;
					key = field.scrollVertical ? "RollDown" : "RollRight";
					break;
				case "dir2":
					if (field.scrollPosition + field.height >= field.scrollRows) return;
					field.scrollAmount = 1;
					key = field.scrollVertical ? "RollUp" : "RollLeft";
					break;
				case "pageDir2":
					if (field.scrollPosition + field.height >= field.scrollRows) return;
					field.scrollAmount = 0;
					key = field.scrollVertical ? "RollUp" : "RollLeft";
					break;
			}
			AXES.Terminal.doPost(AXES.Keyboard.aidKeys[key]);
		},
		_handleInput: function(e)
		{
			var field = e.data._field, val = $(this).val();
			if (field.type == AXES.Terminal.Field.Types.INPUT) 
			{
				this._noFieldValueChangedEvent = true;
				if (val != e.data._masterVal || val != field.getValue())
				{
				    field.setValue(val);
				}
			}
			if (field.ctd) 
			{
				var firstFld = field.getFirstInCTDSeries();
				if (firstFld.focusAttribute) 
				{
					var tmp = firstFld;
					$(tmp.terminalDiv).removeClass(firstFld.focusAttribute);
					while (tmp.ctd_to) 
					{
						var tmp = AXES.Terminal.getFieldById("ifld" + tmp.ctd_to);
						$(tmp.terminalDiv).removeClass(firstFld.focusAttribute);
					}
				}
			}
			else if (field.focusAttribute) 
			{
				$(field.terminalDiv).removeClass(field.focusAttribute);
			}
		},
		_handleFocus: function(e)
		{
			var field = e.data._field;
			if (field.ctd) 
			{
				var firstFld = field.getFirstInCTDSeries();
				if (firstFld.focusAttribute) 
				{
					var tmp = firstFld;
					$(tmp.terminalDiv).addClass(firstFld.focusAttribute);
					while (tmp.ctd_to) 
					{
						tmp = AXES.Terminal.getFieldById("ifld" + tmp.ctd_to);
						$(tmp.terminalDiv).addClass(firstFld.focusAttribute);
					}
				}
			}
			else if (field.focusAttribute) 
			{
				$(field.terminalDiv).addClass(field.focusAttribute);
			}
			if (e.data.offsetOnFocus != undefined) 
			{
				AXES.Layout.setFieldSelection(this, e.data.offsetOnFocus);
				delete e.data.offsetOnFocus;
			}
			else 
			{
				field.setCursor();
				if (this.select) this.select();
			}
			e.preventDefault();
		},
		_handleKeyDown: function(e)
		{
			e.data._IMEActive = (e.which == 229);
			
			var val = $(this).val(), field = e.data._field;
			if (e.axKeyInfo && e.axKeyInfo.mappedTo) 
			{
				switch (e.axKeyInfo.mappedTo) {
					case "FieldMinus":
					case "FieldPlus":
					case "FieldExit":
						if (!field.ctd)
						{
						    // Set to negative and do a field exit.
						    val = field.doFieldKeyFormat(val, e.axKeyInfo.mappedTo);
						    if (val != undefined)
						    {
						        field.setValue(val);
						    	// Change the mapped key so that Axes can handle the
						    	// Field Exit step.
						    	e.axKeyInfo.mappedTo = "FieldExit";
						    	AXES.Keyboard.keydownHandler.call(this, e);
						    }
						}
						// Stop the browser treating it as a normal keypress
						e.preventDefault();
						break;
					case "EraseEOF":
						var sel = AXES.Layout.getFieldSelection(this);
						field.setValue(this.value.substring(sel.start));
						var nextField = field;
						while (nextField.ctd_to) {
							nextField = AXES.Terminal.getFieldById("ifld" + nextField.ctd_to);
							nextField.setValue("");
						}
						e.preventDefault();
						e.axKeyInfo.preventKeypress = true;
						break;
					case "EraseInput":
						var term = AXES.Terminal;
						for (var i = 0, fld; fld = term.sequencedFields[i]; i++) {
							fld.resetValue();
						}
						var home = term.cursorHome;
						term.setCursorPosition(home);
						var focusFld = term.getFieldAtPosition(home);
						if (focusFld) {
							focusFld.setFocus();
						}
						e.preventDefault();
						e.axKeyInfo.preventKeypress = true;
						break;
				}
			}
			else 
			{
				switch (e.which) {
					case 37:
						//Left arrow
						break;
					case 39:
						// Right arrow
						var sel = AXES.Layout.getFieldSelection(this);
						if (sel.start == sel.end && sel.end == this.value.length) {
							// Trying to right arrow past end.
							if (field.ctd_to) 
							{
								nextField = AXES.Terminal.getFieldById("ifld" + nextField.ctd_to);
								nextField.setFocus(0);
							}
						}
						break;
						
				}
			}
		},
		_handleKeyPress: function(e)
		{
			// iPhone/iPad sends a keyPress for backspace (other browsers send no event)
			// additional fix for firefox that validate keypress on numeric field as invalid chars
			if (e.which in [8, 9, 16, 17, 18, 19, 20, 27, 33, 34, 35, 36, 37, 38, 39, 40, 45, 46, 144, 145]) return;
			
			var field = e.data._field;
			if (field instanceof AXES.Terminal.Field.InputField) 
			{
				var theChar = String.fromCharCode(e.which);
				if (field.isMonocase) 
				{
					theChar = AXES.Utils.monocaseString(theChar);
					e.which = theChar.charCodeAt(0);
				}
				if (field.stringHasInvalidChars(theChar)) 
				{
					e.preventDefault();
				}
				field.mdt = true;
			}
		},
		_handleKeyUp: function(e)
		{
			if (e.data._IMEActive && (e.which != 13)) return;
			var field = e.data._field;
			if (e.axKeyInfo && e.axKeyInfo.mappedTo && !e.data._IMEActive) 
			{
				// All mapped keys are either AID keys, cursor control or some
				// special terminal function. Auto Enter (and similar functions)
				// are only triggered by data keys.
				// TODO: Exception to this may be DUP. Need to check and impliment.
			}
			else if (field instanceof AXES.Terminal.Field.InputField) 
			{
				/**
				 * Do not auto enter when the following keys are pressed
				 * 8:	Backspace
				 * 9:	Tab
				 * 16:	Shift
				 * 17:	Ctrl
				 * 18:	Alt
				 * 19:	Pause Break
				 * 20:	Caps Lock
				 * 27:	Esc
				 * 33:	Page Up
				 * 34:	Page Down
				 * 35:	End
				 * 36:	Home
				 * 37:	Left Arrow
				 * 38:	Up Arrow
				 * 39:	Right Arrow
				 * 40:	Down Arroww
				 * 45:	Insert
				 * 46:	Delete
				 * 144:	Num Lock
				 * 145:	Scroll Lock
				 */
				var keys = [8, 9, 16, 17, 18, 19, 20, 27, 33, 34, 35, 36, 37, 38, 39, 40, 45, 46, 144, 145];
				var keyStr = keys.toString();
				var val = $(e.target).val();
				var hostStrInfo = field.checkLength(val);
				if (keyStr.indexOf(e.which) == -1 && hostStrInfo.length >= field.len) 
				{
					var sel = AXES.Layout.getFieldSelection(e.target);
					if (sel.start == sel.end && sel.end == val.length) 
					{
						if (field.isAutoEnter && !field.isFieldExitRequired) 
						{
							AXES.Terminal.doPost(AXES.Keyboard.aidKeys["PostData"]);
						}
						else if (field.ctd_to) 
						{
							var nextID = "ifld" + field.ctd_to;
							var nextField = AXES.Terminal.getFieldById(nextID);
							var focusOffset = undefined;
							if (field.wrap) 
							{
								var i = val.lastIndexOf(" ", hostStrInfo.val.length - 1);
								if (i >= 0) 
								{
									field.setValue(val.substring(0, i));
									nextField.setValue(val.substring(i + 1) + nextField.getValue());
									focusOffset = val.substring(i + 1).length;
								}
							}
							nextField.setFocus(focusOffset);
						}
						else if (!field.isFieldExitRequired) 
						{
							// FER (Field Exit Required) is indicated by (ffw & 0x40)
							// But, Client Access also requires Field Exit if right adjust
							// is set. This is if byte 2, bits 5-7 of the FFW are 101 or 110
							var fer = (field.fieldFormatWord & 0x5) || (field.fieldFormatWord & 0x6);
							if (!fer) AXES.Layout.nextField();
						}
					}
				} else if (e.which == 8) { //handle backspace for numeric field
					var signedNum = (((field.fieldFormatWord & 0x700) >> 8) == 7);
					if (field.rightAlign || signedNum) {
						var i = val.lastIndexOf(" ");
						if (i > 0) {
							field.setValue(val.substring(0, i-1));
						}
					}
				}                
			}
		},
		_handleOutputDblClick: function(e)
		{
			var self = e.data;
			var field = self._field;
			field.setCursor();
			if (field instanceof AXES.Terminal.Field.TextField) 
			{
				var aidKey = self.getPropertyValue("AIDKeyOnDoubleClick");
				if (aidKey) 
				{
					AXES.Terminal.doPost(AXES.Keyboard.aidKeys[aidKey]);
					e.preventDefault();
					this.blur();
					if (document.selection && document.selection.empty) {
						document.selection.empty();
					} else if (window.getSelection) {
						var sel = window.getSelection();
						sel.removeAllRanges();
					}
				}
			}
		}
	}
	AXES.Extensions.add(definition);
}
DEFINE_EXTENSION_axDefault();

function DEFINE_EXTENSION_axNoop()
{
	var definition = {
		/**
		 * Programmatic class name of this extension. It's not visible to end user
		 **/
		name: "axnoop",
		group: "vis",
		/**
		 * Type of object this extension is applicable. Can be either "element" or "screen"
		 * If this extension applies to both element & screen, we put both in an array like so:
		 * type: ["element", "screen"]
		 **/
		type: "autogui",
		
		/**
		 * End-user visible name of this extension
		 **/
		display: TEXT.dev("Do Nothing"),
		
		properties: {},
		
		init: function(field, fieldDiv)
		{
		},
		render: function()
		{
		},
		destroy: function()
		{
		}
	}
	AXES.Extensions.add(definition);
}
DEFINE_EXTENSION_axNoop();

function DEFINE_EXTENSION_ax5250MenuItem()
{
	var definition = {
		/**
		 * Programmatic class name of this extension. It's not visible to end user
		 **/
		name: "ax5250menu",
		group: "vis",
		/**
		 * Type of object this extension is applicable. Can be either "element" or "screen"
		 * If this extension applies to both element & screen, we put both in an array like so:
		 * type: ["element", "screen"]
		 **/
		type: "autogui",
		
		/**
		 * End-user visible name of this extension
		 **/
		display: TEXT.dev("AutoGUI Menu item"),
		
		properties: {
			value: {
				type: "Integer"
			},
			displayText: {
				type: "String"
			},
			useImage: {
				type: "Boolean",
				defaultStatic: false,
				description: TEXT.dev("Replace the menu number with an image.")
			},
			imageURL: {
				type: "String",
				defaultStatic: "/ts/skins/images/menuitem.gif"
			}
		},
		
		init: function(field, fieldDiv)
		{
			this._field = field;
			this._fieldDiv = fieldDiv;
			this._content = null;
			var newEl = null;
			
			var newEl = document.createElement("span");
			newEl.className = "ax-terminal-5250menuitem";
			this._content = $(newEl);
			
			if (this._content && (this._content.length > 0)) 
			{
				this._fieldDiv.appendChild(this._content[0]);
			}
		},
		render: function()
		{
			this._content.empty();
			var link = document.createElement("a");
			link.href = "#";
			$(link).bind("click", this, this._handleClick);
			
			var valStr = this.getPropertyValue("value");
			var sp1 = document.createElement("span");
			sp1.className = "axMenuValue";
			sp1.style.width = (valStr.length * AXES.Layout.characterSize.width) + "px";
			sp1.innerHTML = valStr.replace(/\s/g, "&nbsp;");
			if (this.getPropertyValue("useImage")) 
			{
				sp1.style.visibility = "hidden";
				this._content[0].appendChild(sp1);
				var img = document.createElement("img");
				img.src = this.getPropertyValue("imageURL");
				link.appendChild(img);
			}
			else 
			{
				link.appendChild(sp1);
			}
			var sp2 = document.createElement("span");
			sp2.innerHTML = this.getPropertyValue( "displayText" ).replace( /</g,"&lt;" );
			link.appendChild(sp2);
			
			this._content[0].appendChild(link);
		},
		destroy: function()
		{
			delete this._field;
			delete this._fieldDiv;
		},
		_handleClick: function(e)
		{
			var self = e.data;
			if ( self.isDesignMode() )
				return;
				var ch = AXES.Terminal.cursorHome;
				AXES.Terminal.setCursorPosition(ch);
				var fld = AXES.Terminal.getFieldAtPosition(ch);
			if ( fld )
				fld.setValue( self.getPropertyValue( "value" ).trim() );
				AXES.Terminal.doPost(AXES.Keyboard.aidKeys["Enter"]);
				e.preventDefault();
		}
	}
	AXES.Extensions.add(definition);
}
DEFINE_EXTENSION_ax5250MenuItem();

function DEFINE_EXTENSION_axScreenSize()
{
	AXES.Extensions.definePropertyType("axZoomValue", {
		defaultStatic: 0
	});
	
	var definition = {
		name: "axscrsz",
		display: TEXT.dev("axScreenSize.display"),
		type: "screen",
		helpID: "axScreenSize",
		allowCustomise: true,
		properties: {
			width: {
				type: "Integer",
				defaultStatic: 0,
				description: TEXT.dev("axScreenSize.width")
			},
			height: {
				type: "Integer",
				defaultStatic: 0,
				description: TEXT.dev("axScreenSize.height")
			},
			zoom: {
				type: "axZoomValue",
				description: TEXT.dev("axScreenSize.zoom")
			},
			lockZoom: {
				type: "Boolean",
				defaultStatic: false,
				description: TEXT.dev("axScreenSize.lockzoom")
			}
		},
		init: function(element, htmlElement)
		{
			var zoomVal = Number(this.getPropertyValue("zoom")), zoomLock = this.getPropertyValue("lockZoom"), axesLayout = AXES.Layout, oldZoom;
			
			axesLayout.autoZoomOverride = {
				width: this.getPropertyValue("width"),
				height: this.getPropertyValue("height"),
				zoom: zoomVal,
				lockZoom: zoomLock
			}
			
			switch (zoomVal) {
				case 0:
					break;
				case -2:
					this._oldZoomSetting = axesLayout._zoomSetting;
					axesLayout._zoomSetting = "fit";
					break;
				case -1:
					this._oldZoomSetting = axesLayout._zoomSetting;
					axesLayout._zoomSetting = "auto";
					break;
				default:
					val = zoomVal / 100;
					this._oldZoomSetting = axesLayout._zoomSetting;
					oldZoom = this._oldZoom = axesLayout._zoom;
					axesLayout._zoomSetting = "";
					axesLayout._zoom = (val < 0.1) ? 0.1 : ((val > 5) ? 5 : val);
					window.setTimeout(function()
					{
						if (oldZoom != axesLayout._zoom) 
						{
							axesLayout.fire("ZoomChanged");
						}
					}, 1);
			}
		},
		destroy: function()
		{
			var axesLayout = AXES.Layout, azo = axesLayout.autoZoomOverride;
			var fireChangedEvent = (azo) ? azo.lockZoom : false;
			
			if (this._oldZoomSetting != undefined) 
			{
				if (this._oldZoomSetting != axesLayout._zoomSetting) 
				{
					axesLayout._zoomSetting = this._oldZoomSetting;
					fireChangedEvent = true;
				}
				delete this._oldZoomSetting;
			}
			if (this._oldZoom != undefined) 
			{
				if (this._oldZoom != axesLayout._zoom) 
				{
					axesLayout._zoom = this._oldZoom;
					fireChangedEvent = true;
				}
				delete this._oldZoom;
			}
			
			delete axesLayout.autoZoomOverride;
			if (fireChangedEvent) 
			{
				axesLayout.fire("ZoomChanged");
			}
		}
	};
	
	AXES.Extensions.add(definition);
}
DEFINE_EXTENSION_axScreenSize();
