/*********************************************************************
 *                                                                   *
 * Copyright (C) LANSA Group. All rights reserved                    *
 *                                                                   *
 *********************************************************************/

/**
 * @fileOverview Contains the object and methods for managing the
 * virtual terminal. This is mainly the {@link AXES.Terminal} and 
 * {@link AXES.Terminal.Field} namespaces.
 */

/**
 * Handles all aspects of managing the virtual terminal.  Processes
 * transactions to/from the server and maintains a virtual representation of
 * the terminal.
 * @namespace
 * @borrows AXES.EventsBase.addListener as this.addListener
 * @borrows AXES.EventsBase.removeListener as this.removeListener
 * @borrows AXES.EventsBase.fire as this.fire
 */
AXES.Terminal = {
	title: "",
	key: "",
	CCSID: "-1",
	lk1Assigned : null, /* The lk1 value assigned to this session */
	lk2Assigned : null, /* The lk2 value assigned to this session */
	 /** Value set by the 'Monocase' configuration file directive. Used by
	 * {@link AXES.Utils.monocaseString}.
	 */
	monocaseFlags: 0,
	/** Indicates the type of password. 0=case insensitive, >0=case sensitive */
	passwordLevel: 0,
	signatureModel: 2,
	specialDBCSChars : null, /* Special DBCS characters less with less than 256 unicode value - note this value is destroyed when used */ 
	systemName: "",
	deviceName: "",
	jobUser: "",
	jobNum: "",
	dspf: {},
	/** Array of all fields */
	fields: [],
	/** 
	 * @property {AXES.Terminal.Windows[]} windows An array of {@link AXES.Terminal.Window} objects. There is always at least one item in the array (representing the background).
	 */
	windows: [],
	/** Array of all fields sorted by their position in the terminal */
	sortedFields: [],
	/** Array of input fields in the sequence order defined by the host */
	sequencedFields: [],
	recordFormats: [],
	messages: [],
	htmlItems: [],
	/** Indicates if the terminal message light is on */
	messageLight: false,
	writeFlags: 0,
	readFlags: 0,
	postAction: 0,
	cursorHome: {row: 1, col: 1},
	/** Represents the terminal cursor */
	cursor: {
		row: 1,
		col: 1,
		set: function(newPos)
		{
			if (isNaN(newPos.row) || isNaN(newPos.col)) return;
			if (newPos.row <= AXES.Terminal.rows && newPos.row > 0) this.row = Number(newPos.row);
			if (newPos.col <= AXES.Terminal.cols && newPos.col > 0) this.col = Number(newPos.col);
			AXES.Layout.Cursor.update();
		},
		isInField: function(field)
		{
			return ((this.row >= field.row) && (this.row < field.row + field.height) &&
			(this.col >= field.col) &&
			(this.col < field.col + field.width));
		}
	},
	rows: 25,
	cols: 80,
	columns: 80,
	messageRow: 25,
	busy: true,
	isLogin: false,		// used by RAMP & LongRange too.
	_loginInfo: {},
	isConnected: false,
	isScreenRendered: false,
	transactionLog: [],
	matchedDefinition: null
};
AXES.Terminal.init = function() {
	AXES.Tx.init();
	
	/**
	 * Indicates that a new host screen has arrived. This event is fired prior to the screen being rendered
	 * in the UI.  Returning false will prevent the screen from being rendered.
	 * @name AXES.Terminal#ScreenArrivedPreRender
	 * @event
	 * @param {Object} e Event Parameters
	 * @param {AXES.Terminal} e.terminal The AXES.Terminal object
	 */
	/**
	 * Indicates that a new screen has arrived and has been rendered. This will not fire if rendering has been 
	 * prevented by a AXES.Terminal#ScreenArrivedPreRender handler
	 * @name AXES.Terminal#ScreenArrived
	 * @event
	 * @param {Object} e Event Parameters
	 * @param {AXES.Terminal} e.terminal The AXES.Terminal object
	 */
	/**
	 * Indicates that the screen is about to be posted to the server. Return false to prevent the post from completing.
	 * If you wish to just delay the post while an asynchronous operation is completed, use the 
	 * AXES.Terminal.placeHoldOnPost and AXES.Terminal.clearHoldOnPost methods instead of returning false.
	 * @name AXES.Terminal#ScreenWillPost
	 * @event
	 * @param {Object} e Event Parameters
	 * @param {AXES.Terminal} e.terminal The AXES.Terminal object
	 * @param {AXES.Keyboard.aidKeys} e.aidKey The AID key that the screen will be posted with.
	 */
	/**
	 *
	 * @name AXES.Terminal#ScreenPosted
	 * @event
	 * @param {Object} e Event Parameters
	 * @param {AXES.Terminal} e.terminal The AXES.Terminal object
	 */
	/**
	 * Indicates that aXes was unable to post the screen. Return false from your handler to suppress the default alert.
	 * @name AXES.Terminal#ScreenPostFailed
	 * @event
	 * @param {Object} e Event Parameters.
	 * @param {AXES.Terminal} e.terminal The AXES.Terminal object.
	 * @param {AXES.Keyboard.aidKeys} e.aidKey The AID key that the screen was posted with.
	 * @param {String} e.reason The reason for the failure.
	 */
	/**
	 * Indicates that the host is busy and taking a long time to respond. There are two types of busy message. One that is just informational and one that gives you the option to continue waiting or cancel.
	 * In the case of the latter, e.callback will reference a function that you must call to indicate the choice. In the case of the former, e.callback will be null.
	 * Return false from your handler to suppress the default handling of this message.  If you return false you <strong>must</strong> respond to the allback (if present) otherwise the terminal will remain in a permanently locked state.
	 * @name AXES.Terminal#HostBusy
	 * @event
	 * @param {Object} e Event Parameters.
	 * @param {AXES.Terminal} e.terminal The AXES.Terminal object.
	 * @param {Function} e.callback The callback function to call (passing in true/false) to indicate that you want to continue waiting. For some busy messages, aborting is not an option and this property will be null.
	 */
	/**
	 * Indicates that a message has been received from the Axes TS server.
	 * @name AXES.Terminal#ServerMessage
	 * @event
	 * @param {Object} e Event Parameters.
	 * @param {AXES.Terminal} e.terminal The AXES.Terminal object.
	 * @param {String} e.message The message sent by the server.
	 */
	jQuery.extend(true, this, AXES.EventsBase, {
		_eventNames: {
			ScreenArrivedPreRender: true,
			ScreenArrived: true,
			ScreenWillPost: true,
			ScreenPosted: true,
			ScreenPostFailed: true,
			HostBusy: true,
			ServerMessage: true
		}, _listeners: {}
	});
};
/**
 * (Rodel 5/03/2014) 
 * used by LongRange and RAMP applications 
 * - so their semantic/parameters or behaviours should not be modified without proper impact analysis. 
 */
AXES.Terminal.reset = function() {
	this.postAction = 0;
	this.adminView = false;
	
	this.fields.length = 0;
	this.fields.indexById = {};
	this.fields.indexByName = {};
	this.sortedFields.length = 0;
	this.sequencedFields.length = 0;
	this.recordFormats = [];
	this.windows = [new AXES.Terminal.Window()];
	this.htmlItems.length = 0;
	this.messageLight = false;
	
	this.fKeyMask = "111111111111111111111111";
	this.cursorHome = {
		row: 1,
		col: 1
	};
	this.cursor.set({
		row: 1,
		col: 1
	});
	this.dspf = {};
	
	this.transactionLog.length = 0;
	
	//Create login fields
	var fldTemplate = {
		type: "IField",
		obj: "fld",
		col: 53,
		height: 1,
		width: 10,
		len: 20,
		byp: false,
		fieldFormatWord: 0,
		data: "",
		initialData: "",
		getServerString: function() {
			return this.data.length + "|" + this.data.toUpperCase() + "|";
		}
	};
	
	var i = 0;
	var self = this;
	var addField = function(i, obj, urlParam) {
		var fld = jQuery.extend(new AXES.Terminal.Field.InputField(), fldTemplate, obj);
		fld.symbolicName = fld.id;
		self.fields.indexById[fld.id] = self.fields.indexByName[fld.id] = i;
		self.fields[i] = self.sortedFields[i] = self.sequencedFields[i] = fld;
		var val = "";
		if (urlParam) val = AXES.urlParameters[urlParam];
		if (self._loginInfo[fld.id]  && (fld.id != "loginPwd")) val = self._loginInfo[fld.id];
		if (val) fld.setValue(val);
		
		return fld;
	}
	
	var userFld = addField(i++, {id: "loginUser", row: 6}, "user");
	var pwdFld = addField(i++, {id: "loginPwd", row: 7}, "pwd");
	if (this.passwordLevel > 0) 
	{
		pwdFld.getServerString = function()
		{
			return this.data.length + "|" + this.data + "|";
		}
	}
	
	addField(i++, {id: "loginDevice", row: 9}, "device");
	addField(i++, {id: "loginPgm", row: 10}, "program");
	addField(i++, {id: "loginMenu", row: 11}, "menu");
	addField(i++, {id: "loginCurlib", row: 12}, "curlib");
	if (!this._loginInfo["loginScreenType"]) this._loginInfo["loginScreenType"] = this.screenType;
	addField(i++, {id: "loginScreenType", row: 8}, "screentype");
	addField(i++, {id: "loginReconnect", row: 13, width: 1}, "reconnect");

	if (AXES.urlParameters["encrypt"] && AXES.urlParameters["pwd"]) 
	{
		//decrypt password created by Command.html in the development kit	
		var bpi = "31415926535897932384626433832795028841971693993751058209749445923078164062";
		pwdFld.setValue(ARC4decrypt(bpi, AXES.urlParameters["pwd"], MD5s));
	}

	this.isLogin = true;
	this.isConnected = false;
	
};
AXES.Terminal.processInitInstructions = function(xml) {
	var node = xml.find("Connect");
	if (node.length > 0) 
	{
		this.CCSID = node.attr("ccsid");
		this.isDBCSCapable = (node.attr("cscaps") == "3");
		this.screenType = node.attr("wkstn");
		this.monocaseFlags = parseInt(node.attr("monocase"));
		AXES.Layout.AutoGUI.enabled = AXES.Layout.AutoGUI.autoGUIOn = Boolean(parseInt(node.attr("autogui")));
		this.passwordLevel = parseInt(node.attr("pwdlvl"));
		this.signatureModel = parseInt(node.attr("signaturemodel"));
		this.specialDBCSChars = node.attr("specialDBCSChars"); 
		this.systemName = node.attr("sysname");
		var theme = node.attr("TS2defaulttheme");
		var allowThemeChange = node.attr("TS2changetheme");
		if (allowThemeChange == undefined) allowThemeChange = 1;
		AXES.UI._allowThemesMenu = Boolean(parseInt(allowThemeChange));
		if (('localStorage' in window) && window['localStorage'] !== null) 
		{
			if (AXES.UI._allowThemesMenu) 
			{
				theme = localStorage['axTheme'] || theme;
			}
			else 
			{
				if (localStorage['axTheme'] != undefined) delete localStorage['axTheme'];
			}
		}
		if (!theme) theme = "theme_default";
		AXES.UI.setTheme(theme);
	}
	this.busy = false;
};

AXES.Terminal.getToolConnectionInformation = function()
{
	if (this.isConnected) 
	{
		return {
			usable: true,
			profile: this.jobUser,
			password: this._loginInfo.loginPwd || "",
			jobnumber: this.jobNum
		}
	}
	else 
	{
		return {
			usable: false,
			profile: "",
			password: "",
			jobnumber: ""
		}
	}
};

AXES.Terminal.transactionArrived = function(xml) {
	this.isScreenRendered = false;
	this.busy = false;
	AXES.Layout.setBusy(false);
	this.isLogin = false;
	this.isConnected = true;
	this.messages.length = 0;

	var $xml = $(xml);
	this.deviceName = $xml.attr("device_name");
	this.jobUser = $xml.attr("jobuser");
	this.jobNum = $xml.attr("jobnum");
	
    var xmltempattr = $xml.attr("lk1");
    if (xmltempattr) /* Only the first transaction has these and they are always in pairs */  
    {
       AXES.Terminal.lk1Assigned = xmltempattr;
       xmltempattr = $xml.attr("lk2");  
       if (xmltempattr) AXES.Terminal.lk2Assigned = xmltempattr;
    }   
    			
	var continueNode = null;
	var serverMessage = "";
	var dirty = false;
	var verifyRequest = null;
	
	var self = this;
	$xml.children().each(function(index, node) {
		var $node = $(node);
		var msg = $node.attr("alert");
		if (msg) self.messages.push(msg);
		switch (node.nodeName) {
			case "LOGIN":
				var dataString = "";
				for (var i = 0, fld; fld = self.sortedFields[i]; i++) {
					if (fld.id != "loginReconnect") {
						dataString += fld.getServerString();
						self._loginInfo[fld.id] = fld.getValue();
					}
				}
				AXES.Tx.aidKey = $node.attr("id");
				AXES.Tx.url = "connect";
				AXES.Tx.command = "OPEN";
				AXES.Tx.data = dataString;
				AXES.Tx.sendLogin($node.attr("nonce"));
				self.busy = true;
				AXES.Layout.setBusy(true);
				// return now to prevent normal screen update
				return false;
			case "COMMAND":
				self.processCOMMAND(node);
				dirty = true;
				break;
			case "VIEW":
				this.adminView = {
					userName: $node.attr("user"),
					userRef: $node.attr("ref"),
					userDevice: $node.attr("device"),
					jobNum: $node.attr("jobnum"),
					systemName: $node.attr("sysname")
				};
				// TODO: The VIEW node indicates that this is an admin view (watching
				// another users session.  Apart from processing the tag and preventing
				// posts, TS2 does not yet implement Admin view.
				break;
			case "DISCONNECT":
				self.reset();
				msg = $node.attr("msg");
				if (msg) {
					// The "alert" attribute already generates a message.
					// The msg/hdr combo here appears to have been intended to
					// produce an actual dialog. TS1 doesn't do that so, for 
					// now at least, we'll do the same and just log a message.
					msg = $node.attr("hdr") + " - " + msg;
					self.messages.push(msg);
				}
				dirty = true;
				break;
			case "VERIFY":
				verifyRequest = {hdr: $node.attr("hdr"), msg: $node.attr("msg")};
				break;
			case "RESTART":
				self.reset();
				msg = $node.attr("msg");
				if (msg) {
					// The "alert" attribute already generates a message.
					// The msg/hdr combo here appears to have been intended to
					// produce an actual dialog. TS1 doesn't do that so, for 
					// now at least, we'll do the same and just log a message.
					msg = $node.attr("hdr") + " - " + msg;
					self.messages.push(msg);
				}
				dirty = true;
				break;
			case "ERROR":
				var msg = TEXT.Core("Error received from the aXesTS server: %s1", $(node).attr("msg"));
				AXES.onError(1, msg);
				break;
			case "MESSAGE":
				// Queue messages to display after screen rendered.
				serverMessage = $node.attr("msg").replace(/[ \f\n\r\t]*$/,""); // strip off trailing spaces
				break;
			case "TIMEOUT":
			case "CONTINUE":
				continueNode = node;
				break;
			case "MessageWait":
				self.messageLight = true; 
				break;  
			default:
				if (node.nodeType == NODE_ELEMENT) 
				{
					var msg = TEXT.Core("Error in transaction from server: Unknown transaction item (%s1).", node.nodeName);
					AXES.onError(1, msg);
				}
		}
	});
	
	this.transactionLog.push(xml);
	if (this.busy) return;

	// Retrieve extended data from the server
	if (AXES.ExtendedData) AXES.ExtendedData.open();
	
	//
	if (dirty) {
		this.updateForNewScreen();
	}
	
	if (verifyRequest) {
		this._verifyUser(verifyRequest);	
	}
	if (serverMessage) {
		AXES.Layout.displayServerMessage(serverMessage);
	}
	if (continueNode) 
	{
		var $c = $(continueNode);
		var continueType = $c.attr("type");
		if (continueType && continueType == "5") return; // Ignore type 5
		
		var silent = $c.attr("silent");
		var callback = null;
		var done = false;
	
		this.postAction = 0;
		if (silent) 
		{
			this.doSysPost("CONTINUE");
		}
		else 
		{
			/** @ignore */
			callback = function(keepWaiting)
			{
				// If multiple event handlers end up responding to this,
				// only the first is processed.
				if (done) return;
				done = true;
				if (keepWaiting) 
				{
					AXES.Terminal.doSysPost("CONTINUE");
				}
				else 
				{
					AXES.Layout.setBusy(false);
				}
			}
		}
		
		if (this.fire("HostBusy", [{
			terminal: this,
			callback: callback
		}])) {
			if (callback) AXES.Layout.displayHostBusyMessage(callback);
		}
		
	}
};

/**
 * Get lk1 && lk2 values
 */
AXES.Terminal.getLK12Assignments = function () {
	var o = { lk1: "", lk2: "" };

	if (typeof (AXES.Terminal.lk1Assigned) == "string") o.lk1 = AXES.Terminal.lk1Assigned;
	else o.lk1 = "NA";
	if (typeof (AXES.Terminal.lk2Assigned) == "string") o.lk2 = AXES.Terminal.lk2Assigned;
	else o.lk2 = "NA";

	return o;
};

/**
 * Get terminal's properties such as device name & lk1/2
 */
AXES.Terminal.getGeneralInfo = function ()
{
	var result =
	{
		deviceName: AXES.Terminal.deviceName,
		lk12: AXES.Terminal.getLK12Assignments()
	};
	return result;
};

/**
* Attempts to detect "fake" windows. These are windows drawn on the screen with border
* characters of some sort instead of being specified by a <window> tag in the XML.
* This method detects windows that were defined in the DDS but have been rendered as
* "fake" windows by the OS because space was not provided at either side for attribute
* characters.
*/
AXES.Terminal.fixFakeWindows = function() 
{
	var winRfmt = null, wins = this.windows, win, wl = wins.length, 
		fields = this.fields, fld, fl = fields.length;
		
	if ((AXES.urlParameters.getInt("popups",3) & 1) == 0) return;
	
	for (var i = 0, rf; rf = this.recordFormats[i]; i++)
	{
		if (rf.type == AXES.Terminal.RecordFormat.Types.WINDOW) winRfmt = rf;
	}
	
	if (winRfmt) {
		if (wl == 1)
		{
			// There are no windows, create one
			wins.push(win = new AXES.Terminal.Window(winRfmt));
		} else {
			// There are windows. If the top is a real window then we are done. If it is a fake window
			// then this screen has probably modified it and there may be fields that need moving into it
			win = wins[wl-1];
			if (win.isRealWindow) return;
		}

		var r1 = winRfmt.startLine,r2 = winRfmt.endLine,c1 = winRfmt.startCol,c2 = winRfmt.endCol;
		for (var i = 0; i < fl; i++) 
		{
			var fld = fields[i];
			if (fld) 
			{
				if (fld.win) continue;
				var r = fld.row, c = fld.col;
				if ((r >= r1) && (r <= r2) && (c >= c1) && (c <= c2)) 
				{
					if ((r == r1) || (r == r2) || (c == c1) || (c == c2)) 
					{
						// It is part of the border
						fld.data = fld.initialData = "";
					} else {
						fld.win = win.id;
						fld.winNum = win.idNum;
					}
				} 
			}
		}
	}
}
AXES.Terminal.updateForNewScreen = function()
{
	if (this.isLogin) 
	{
		this.screenLoaded();
		return;
	}
	
	this.fixFakeWindows();
	
	this.windows.indexById = {};
	for (var i = 0, win; win = this.windows[i]; i++) 
	{
		this.windows.indexById[win.id] = i;
		win.fields.length = 0;
	}
	
	var newFlds = [];
	newFlds.indexById = {};
	newFlds.indexByName = {};
	this.sortedFields.length = 0;
	this.sequencedFields.length = 0;
	var l = this.fields.length;
	var j = 0;
	for (var i = 0; i < l; i++) 
	{
		var fld = this.fields[i];
		if (fld) 
		{
			newFlds[j] = this.sortedFields[j] = fld;
			newFlds.indexById[fld.id] = j;
			j++;
		}
	}
	this.fields = newFlds;
	this.sortedFields.sort(function(a, b)
	{
		if (a.winNum == b.winNum) 
		{
			if (a.row == b.row) 
			{
				return (a.col - b.col);
			}
			else 
			{
				return (a.row - b.row);
			}
		}
		else 
		{
			return (a.winNum - b.winNum);
		}
	});
	
	if (this.postAction == 5) {
		this._aidKeyForPost = 0;
		this.completeDoPost();
		return;
	}
	
	var defaultRF = null;
	for (var j = this.recordFormats.length-1; j >= 0; j--) {
		var rf = this.recordFormats[j];
		if ((defaultRF == null) && (rf.startLine == 0) && (rf.endLine == 0)) {
			defaultRF = rf;
		}
		rf.winNum = 0;
		for (var i = this.windows.length-1; i>0; i--) 
		{
			win = this.windows[i];
			if (rf.startLine >= win.row && rf.endLine <= (win.row + win.height + 1)) {
				rf.winNum = win.idNum;
				break;
			}
		}
	}
	
	for (var i = 0, fld; fld = this.sortedFields[i]; i++) 
	{
		// Add the field to the appropriate window
		var winIndex = (fld.win) ? this.windows.indexById[fld.win] : 0;
		this.windows[winIndex].fields.push(fld);
	}
	
	// The sequencedFields array is used to control default tab order
	// and contains only input and select fields in the top-most window.
	var win = this.windows[this.windows.length - 1];
	for (var i = 0, fld; fld = win.fields[i]; i++) 
	{
		if ((fld.type == AXES.Terminal.Field.Types.INPUT && !fld.byp) || fld.type == AXES.Terminal.Field.Types.SELECT) 
		{
			this.sequencedFields.push(fld);
		}
		
		// Add the field to the appropriate recordFormat.
		var found = false, winRFType = AXES.Terminal.RecordFormat.Types.WINDOW;
		for (var j = this.recordFormats.length-1; j >= 0; j--) {
			var rf = this.recordFormats[j];
			if ((rf.type != winRFType) && (fld.row >= rf.startLine) && (fld.row <= rf.endLine)) {
				fld.recordFormat = rf;
				rf.fields.push(fld);
				found = true;
				break;
			}
		}
		if (!found && defaultRF) defaultRF.fields.push(fld);
	}

	if (this.recordFormats.length == 0 && !this.isLogin) 
	{
		AXES.SysTrace.warn(TEXT.core("The screen contains no record formats. This usually indicates that the mandatory PTFs have not been applied."));
	}
	
	this.screenLoaded();
};
AXES.Terminal._verifyUser = function(verifyRequest)
{
	var self = this;
	var done = false;
	var callback = function(dataToSend)
	{
		// If multiple event handlers end up responding to this,
		// only the first is processed.
		if (done) return;
		done = true;
		if (dataToSend) 
		{
			dataToSend.user = escape(dataToSend.user.toUpperCase());
			var pwHash = escape((self.passwordLevel == 0) ? dataToSend.password.toUpperCase() : dataToSend.password);
			if (AXES.Tx.secureOn) 
			{
				for (var i = pwHash.length; i < 128; i++) 
					pwHash += '\x5c';
				
				pwHash += dataToSend.user;
				for (var i = pwHash.length; i < 256; i++) 
					pwHash += '\x5c';
				pwHash = MD5s(pwHash + "(NOT RFC2104)");
			}

			var data = dataToSend.user.length + "|" + dataToSend.user + "|";
			data += pwHash.length + "|" + pwHash + "|0||0||0||0||0||0||";
			
			AXES.Tx.command = "VERIFY";
			AXES.Tx.cursorRow = self.cursor.row;
			AXES.Tx.cursorCol = self.cursor.col;
			AXES.Tx.aidKey = "";		
			AXES.Tx.data = data;
			AXES.Tx.sendLoginVerify();
			self.busy = true;
			AXES.Layout.setBusy(true);
		}
		else 
		{
			self.reset();
			self.messages.push(TEXT.Core("Session Suspended"));
			self.updateForNewScreen();
		}
	}

	this._getUserCredsForVerification(verifyRequest.hdr, verifyRequest.msg, this._loginInfo["loginUser"], callback);
}

AXES.Terminal._getUserCredsForVerification = function(hdr, msg, loginInfo, callback) {
	if (!AXES.AppHost) {
		AXES.Layout.displayVerifyUserDialog(hdr, msg, loginInfo, callback);		
		return;
	}

	var onCompleted = function (result)
	{
		if (result.status == "OK")
		{
			var response =
			{
				user: result.user,
				password: result.password
			};

			if (typeof callback == "function") callback(response);
		}
		else if (result.status == "CANCELLED")
		{
			AXES.Layout.displayVerifyUserDialog(hdr, msg, loginInfo, callback);
		}
		else
		{
			alert("Unable to request credentials.\n" + result.message);
		}
	};

	var code = msg.substring(0, 7);
	if (code == "AXS8203")
	{
		AXES.AppHost.exec("Terminal", "requestCredentials", null, onCompleted);
	}
	else
	{
		var args = { message: msg };
		AXES.AppHost.exec("Terminal", "loginFailed", args, null);
	}
};

AXES.Terminal.screenLoaded = function() {
	this.matchedDefinition = AXES.Screens.findMatch(this);
	if (this.matchedDefinition) 
	{
		AXES.SysTrace.info(TEXT.Core("Screen id %s1 (%s2) matched to current screen.", this.matchedDefinition.id, this.matchedDefinition.name));
	}
	else 
	{
		AXES.SysTrace.error(TEXT.Core("No match found for current screen."));
	}
	this.updateWithMatch(this.matchedDefinition);
	
	if (this.fire("ScreenArrivedPreRender", [{ terminal: this }])) 
	{
		this._showLayout = true;
		AXES.Layout.doLayout();
		this.isScreenRendered = true;

		// Extract extended data if required
		if (AXES.ExtendedData) AXES.ExtendedData.extractData();

		// Prevent long winded handlers delaying presentation of the layout
		var term = this;
		window.setTimeout(function()
		{
			term.fire("ScreenArrived", [{
				terminal: term
			}]);
		}, 1);
	} else {
		AXES.Layout.clearLayout();
	}
};
/**
 * Updates the Terminal with details from the supplied screen definition. If no
 * definition is supplied, clears any details currently applied
 * @param {Object} matchedDef The screen definition to apply.
 */
AXES.Terminal.updateWithMatch = function(matchedDef)
{
	var matchedDefRFs;
	if (!this.isLogin) this.fields.indexByName = {};
	
	if (matchedDef) 
	{
		this.symbolicName = matchedDef.name;
		this.variantName = (matchedDef.variantName) ? matchedDef.variantName : "";
		this.description = matchedDef.description;
		
		matchedDefRFs = matchedDef.recordFormats || {};
		
		for (var fldId in matchedDef.fields) 
		{
			var fldDef = matchedDef.fields[fldId];
			if (fldDef.name != "") 
			{
				var adjustedId = fldId, win = this.windows[this.windows.length - 1];
				if (win.isRealWindow) 
				{
					// The field ID encodes the window number.
					// This varies depending on the order windows are opened
					// so it is not part of the definition.  We need to 
					// insert the window number back into the ID before
					// looking for the field in the current screen.
					var match = fldId.match(/(ifld|text)(.*)/);
					if (match.length > 2) 
					{
						var idNum = parseInt(match[2], 16);
						var winNum = (parseInt(win.id.substring(3)) * 2) << 12;
						adjustedId = match[1] + (idNum + winNum).toString(16);
					}
				}
				var fldIndex = this.fields.indexById[adjustedId],
					existingField = this.fields[this.fields.indexByName[fldDef.name]],
					newField = this.fields[fldIndex], addNameToField = true;
				if (fldIndex != undefined) {
					if (existingField !== undefined) {
						// Already exists
						// To keep this simple, we are trying to determine which field is visible to the user.
						// We know that an input filed with a BYP flag is definately hidden. An output field that is
						// empty is not technically hidden but, for all practical purposes it is (and TS1 ignores it).
						var existingScore = 0, newScore = 0;
						if (existingField.byp) existingScore = -3;
						if ((existingField.type == AXES.Terminal.Field.Types.OUTPUT) && (existingField.data == "")) existingScore = -1;
						if (newField.byp) newScore = -2;
						if ((newField.type == AXES.Terminal.Field.Types.OUTPUT) && (newField.data == "")) newScore = 0;
						// The offset in scores ensures that, if both have the same score, we do what we did before and keep the new one.
						if (newScore < existingScore) addNameToField = false
					}
					if (addNameToField) {
						this.fields.indexByName[fldDef.name] = fldIndex;
						this.fields[fldIndex].symbolicName = fldDef.name;
					}
				}
			}
		}
	}
	else 
	{
		this.symbolicName = (this.isLogin) ? "MainLogin" : "";
		this.variantName = "";
		this.description = "";
		matchedDefRFs = {};
	}
	for (var i = 0, rf; rf = this.recordFormats[i]; i++) 
	{
		if (rf.calculateStructure) rf.calculateStructure(matchedDefRFs[rf.name]);
	}
};

/**
 * Returns the signature string for the current screen
 * @param {String} type The type of signature to return. "A" is the current type, anything else returns the 1.3.3 style signature.
 * @return {String} The signature string
 */
AXES.Terminal.getKey = function(type) 
{
	if (this.isLogin) return "";
	var key, winNum = this.windows[this.windows.length - 1].idNum, rf;
	if (this.signatureModel == 1) type = ""; //Settings are forcing old signature.
	switch (type) {
		case "A":
			// Return the new style key (record formats sorted and any
			// sflmsgrcd types removed
			key = "A." + this.dspf.name;
			// Make a copy of the recordformat array and sort it
			var sortedRF = this.recordFormats.slice(0,this.recordFormats.length);
			sortedRF.sort(function(a, b) {
				if (a.name < b.name) return -1;
				if (a.name > b.name) return 1;
				return 0;
			});
			var l = sortedRF.length;
			for (var i=0; i<l; i++) 
			{
				rf = sortedRF[i];
				if (rf.winNum == winNum && sortedRF[i].type != AXES.Terminal.RecordFormat.Types.SUBFILE_MESSAGE_RECORD) {
					key += "." + rf.name;
				}
			}
			break;
		default:
			//Return the old style key
			key = this.dspf.name;
			var l = this.recordFormats.length;
			for (var i=0; i<l; i++) 
			{
				rf = this.recordFormats[i];
				if (rf.winNum == winNum) key += rf.name;
			}
	}
	return key;
}
/**
 * Returns a string containing an HTML table representation of the area
 * of the screen defined by supplied box parameter.
 * @param {Object} box The area of the screen to return. If not supplied, the whole screen is assumed.
 * @param {Number} box.left
 * @param {Number} box.right
 * @param {Number} box.top
 * @param {Number} box.bottom
 */
AXES.Terminal.getAsTable = function(box)
{
	if (!box) box = {
		left: 1,
		right: this.cols,
		top: 1,
		bottom: this.rows
	};
	box.width = box.right - box.left + 1;
	box.height = box.bottom - box.top + 1;
	var a = new Array(box.height);
	for (var i = 0; i < box.height; i++) 
	{
		a[i] = new Array(box.width);
	}
	
	var win = this.windows[0], winNum = 0;
	do 
	{
		for (var i = 0, fld; fld = win.fields[i]; i++) 
		{
			// Does the field intersect the box?
			var fa = fld.getTextArray();
			for (var j = 0; j < fa.length; j++) 
			{
				var fldInfo = {};
				fldInfo.row = fld.row + j + win.row;
				fldInfo.left = fld.col + win.col;
				fldInfo.right = fldInfo.left + fld.width - 1;
				fldInfo.boxRow = fldInfo.row + j - box.top;
				fldInfo.boxLeft = fldInfo.left - box.left;
				fldInfo.boxRight = fldInfo.right - box.left;
				fldInfo.width = fld.width;
				
				if (fldInfo.boxRow >= 0 && fldInfo.boxRow < box.height) 
				{
					if (fldInfo.boxLeft < box.width && fldInfo.boxRight >= 0) 
					{
						// Some part of it is inside the box.
						var s = 0;
						if (fldInfo.boxLeft < 0) {
							s = Math.abs(fldInfo.boxLeft);
							fldInfo.width += fldInfo.boxLeft;
							fldInfo.boxLeft = 0;
						}
						var e = (fldInfo.boxRight >= box.width) ? box.width - fldInfo.boxLeft : fldInfo.width;
						a[fldInfo.boxRow][fldInfo.boxLeft] = fa[j].substring(s, e);
					}
				}
			}
		}
		
		win = this.windows[++winNum];
		if (win) 
		{
			// Check if any existing fields ovelap this window
			// and truncate them appropriately.
			// First convert window dimensions to box co-ordinates
			var winLeft = Math.max(win.left - box.left, 0);
			var winRight = Math.min(win.col + win.width - 1 - box.left, box.right);
			var winTop = Math.max(win.row - box.top, 0);
			var winBottom = Math.min(win.row + win.height - 1 - box.top, box.bottom);
			
			for (var row = winTop; row <= winBottom; row++) 
			{
				var col = 0;
				while (col < box.width) 
				{
					var item = a[row][col];
					if (typeof item == "string") 
					{
						if (item.length == 0) item = a[i][col] = " ";
						var nextCol = col + item.length;
						if (col < winLeft) 
						{
							if (col + item.length >= win.col) 
							{
								// Truncate to window start;
								a[i][col] = item.substring(0, win.col - col);
								if (nextCol >= win.col + win.width) 
								{
									// Extends past the window so add the last bit
									a[i][win.col + win.width] = item.substring(win.col + win.width - col);
								}
							}
						} else if (col < win.col + win.width) 
						{
							// Delete
							a[i][col] = null;
							if (nextCol >= win.col + win.width) 
							{
								// Extends past the window so add the last bit
								a[i][win.col + win.width] = item.substring(win.col + win.width - col);
							}
						}
						col = nextCol;
						continue;
					}
					col++;
				}
			}
		}
	} while (win);
	// Now we have a 2-d array of fields.  Identify columns in which
	// a field starts or ends
	var starts = new Array(box.width);
	var ends = new Array(box.width);
	for (var j = 0; j < box.width; j++) 
	{
		starts[j] = ends[j] = 0;
	}
    var offset;
    for ( var i = 0; i < box.height; i++ )
        for ( var j = 0; j < box.width; j++ )
            if ( typeof a[ i ][ j ] == "string" && "" != a[ i ][ j ] )
            {
                starts[ j ]++;
                offset = a[ i ][ j ].length - 1;
                ends[ j + offset ]++;
                j += offset;
            }
	// Extend each field to the right until it hits a column where
	// another field starts.
	for (var i = 0; i < box.height; i++) 
	{
		for (var j = 0; j < box.width; j++) 
		{
			var item = a[i][j];
			if (typeof item == "string") 
			{
				var k = j + item.length - 1;
				ends[k]--;
				while (!starts[k] && k < box.width) 
					k++;
				a[i][j] = item.rpad(k - j);
				ends[k]++;
			}
		}
	}
    // Now create the html. By assuming a column width of one character
    // and setting the colspan of each cell tot he number of characters of
    // its content we get a reasonable approximation of the screen in
    // table form.
    for ( var i = 0; i < box.height; i++ )
    {
        var s = "<tr><td>";
        var j = 0;
        var gapStart = 0;
        while ( j < box.width )
            if ( typeof a[ i ][ j ] == "string" && "" != a[ i ][ j ] )
            {
                // Fill in any gaps with an empty cell.
                if ( gapStart < j )
                    s += ( '<td colspan="' + ( j-gapStart ) + '"></td>' );
                s += ( '<td colspan="' + a[ i ][ j ].length + '">' + a[ i ][ j ] + '</td>' );
                j += a[ i ][ j ].length;
                gapStart = j;
            }
            else
                j++;
		// Add an empty cell to the end if necessary.
		if (gapStart < j) 
		{
			s += ('<td colspan="' + (j - gapStart + 1) + '"></td>');
		}
		a[i] = s + "</tr>";
	}	
	
	return "<table>" + a.join("\r\n") + "</table>";
}
/**
 * Returns a string containing a text representation of the area
 * of the screen defined by supplied box parameter.
 * @param {Object} box The area of the screen to return. If not supplied, the whole screen is assumed.
 * @param {Number} box.left
 * @param {Number} box.right
 * @param {Number} box.top
 * @param {Number} box.bottom
 */
AXES.Terminal.getAsText = function(box) {
	if (!box) box = {
		left: 1,
		right: this.cols,
		top: 1,
		bottom: this.rows
	};
	
	// Create an array of strings and insert the contents
	// of each field into the right places.
	var a = new Array(this.rows);
	var s = "".rpad(this.cols);
	for (var i=0;i<this.rows;i++) {a[i] = s.split(""); }
	
	var win = this.windows[0];
	var winNum = 0;
	do {
		for (var i = 0, fld; fld = win.fields[i]; i++) 
		{
			var fa = fld.getTextArray();
			for (var j = 0; j < fa.length; j++) 
			{
				var pos = fld.col + win.col - 1;
				var str = fa[j];
				var l = str.length;
				var index = fld.row + j + win.row - 1;
				for (var k = 0; k < l; k++) 
				{
					a[index][pos++] = str.charAt(k);
					if (AXES.Utils.isDBCS(str.charCodeAt(k))) a[index][pos++] = null;
				}
			}
		}
		for (var i=0;i<this.rows;i++) {a[i] = a[i].join(""); }
		win = this.windows[++winNum];
		if (win) 
		{
			// Just in case of overlap, insert spaces where the window is.
			var s = "".rpad(win.width);
			var b = win.row + win.height - 1
			for (var i = win.row - 1; i < b; i++) 
			{
				a[i] = a[i].substring(0, win.col - 1) + s + a[i].substring(win.col + s.length - 1);
			}
		}
	} while (win);
	
	// Now reduce to the requested box size
	a.splice(box.bottom, this.rows - box.bottom);
	a.splice(0, box.top-1);
	for (var i=0;i<a.length;i++) {
		a[i] = a[i].substring(box.left-1,box.right-1);
	}
	return a.join("\r\n");
}
AXES.Terminal.processCOMMAND = function(commandNode) {
	switch ($(commandNode).attr("class")) {
		case "":
			// TS1 recognises 4 commands here. CLEAR, HTML, STYLESHEET, TRANSFORM.
			// All involve direct manipulation of the Axes internals by the server
			// in ways no longer supported (e.g. using XSL transforms).  In addition,
			// there appears to be no code in the server source that generates
			// these node. So, these appear to be no longer used.  Will log them
			// if they occur.
			$(commandNode).children().each(function(index, node) {
				AXES.SysTrace(TEXT.Core("Unknown server command: %s1", node.nodeName));
			});
			break;
		case "5250-DDS":
			AXES.Terminal.DDSCommands.process(commandNode);
			break;
		case "object":
			AXES.Terminal.ObjectCommands.process(commandNode);
			break;
		case "WebFacing":
			// Not supported. TS1 displays an alert for each unsupported
			// command (which happens to be all).  Probably intended to
			// support them but never did. We will just log them.
			$(commandNode).children().each(function(index, node) {
				AXES.SysTrace(TEXT.Core("Unknown WebFace command: %s1", node.nodeName));
			});
			break;
		default:
			//Note: COMMAND is the name of the 5250 instruction
			var msg = TEXT.Core("Error in transaction from server: Unknown COMMAND class (%s1)", cmdNode.attr("class"));
			AXES.onError(2, msg);
			break;
	}
};
/**
 * (Rodel 5/03/2014) 
 * used by LongRange and RAMP applications 
 * - so their semantic/parameters or behaviours should not be modified without proper impact analysis. 
 */
AXES.Terminal.doPost = function(aidKey) {
	if (this.busy) return;
	if (this.adminView) {
		onError(TEXT.Core("The administrator view cannot post data."));
		return;
	}
	this._postHoldKeys = [];
	var ae = AXES.Layout.getActiveElement();
	if (ae && this.isLogin)
	{
		$(ae).blur().trigger("focusout");
		$("#terminalInner")[0].focus();
	}
	this._aidKeyForPost = (typeof aidKey == "number") ? aidKey : AXES.Keyboard.aidKeys[aidKey];
	if (this._aidKeyForPost) 
	{
		if (this.fire("ScreenWillPost", [{terminal: this,aidKey: this._aidKeyForPost}]) == false) {
			this._aidKeyForPost = null;
			return;
		}

		if (AXES.ExtendedData) 
		{
			//TODO: AXES.ExtendedData should subscribe to "ScreenWillPost" and do this itself only if necessary.
			var self = this, holdKey = self.placeHoldOnPost();
			var onCloseCompleted = function() { self.clearHoldOnPost(holdKey); }
			AXES.ExtendedData.close(onCloseCompleted);
		}
		else
		{
			this.completeDoPost();
		}
	}
	else 
	{
		AXES.SysTrace.error(TEXT.core("Post attempted with invalid AID key (%s1)", aidKey));
	}
};
/**
 * Prevents the screen from posting. This can only be called from within a "ScreenWillPost" event handler.
 * This does not prevent a screen from ever being posted. It simply prevents the current post operation
 * from completing. All holds will be cleared and the "ScreenWillPost" event re-sent if the user re-attempts
 * the post.
 * @returns {Object} a key that must be passed to clearHoldOnPost to release the hold.
 */
AXES.Terminal.placeHoldOnPost = function() {
	var holdKey = new Object();
	this._postHoldKeys.push(holdKey);
	return holdKey;
}
/**
 * Releases a hold on the screen post. If the key is still valid (hasn't been invalidated by a new post or a
 * global cancellation of the post) and there are no other holds on the post operation, then the post will be 
 * completed.
 * @param {Object} holdKey The key returned by placeHoldOnPost.
 */
AXES.Terminal.clearHoldOnPost = function(holdKey) {
	for (var i=0,key; key=this._postHoldKeys[i]; i++) {
		if (holdKey === key) {
			// Valid key
			this._postHoldKeys.splice(i,1);
			this.completeDoPost();
			break;
		}
	}
}
/**
 * Finishes the doPost.  doPost sends an event giving listeners the option to
 * cancel the post. Because a JavaScript dialog is asynchronous, the listener
 * may need to cancel the post while waiting for confirmation and then
 * continue with it later.  This allows that to happen.
 */
AXES.Terminal.completeDoPost = function() {
	if (this._aidKeyForPost == null) return;
	if (this._postHoldKeys.length) return;
	
	if (this.isLogin) {
		var fld = this.getFieldById("loginReconnect");
		if (fld && fld.getValue() && fld.getValue().toUpperCase() != "N") 
		{
			if ((fld.id != "loginPwd") && (fld.id != "loginReconnect")) self._loginInfo[fld.id] = fld.getValue();
			var userFld = this.getFieldById("loginUser");
			var userName = userFld.getValue().toUpperCase();
			var pwdFld = this.getFieldById("loginPwd");
			var pwd = (this.passwordLevel == 0) ? pwdFld.getValue().toUpperCase() : pwdFld.getValue();
			var deviceFld = this.getFieldById("loginDevice");
			
			for (var i = 0, fld; fld = this.sortedFields[i]; i++) 
			{
				if ((fld.id != "loginPwd") && (fld.id != "loginReconnect")) this._loginInfo[fld.id] = fld.getValue();
			}
				
			var pwHash = escape(pwd);
			if (AXES.Tx.secureOn) 
			{
				for (var i = pwHash.length; i < 128; i++) 
					pwHash += '\x5c';
				
				pwHash += userName;
				for (var i = pwHash.length; i < 256; i++) 
					pwHash += '\x5c';
				pwHash = MD5s(pwHash + "(NOT RFC2104)");
			}
			
			var data = userName.length + "|" + userName + "|"
				+ pwHash.length + "|" + pwHash + "|0||0||0||0||0||0||";
			
			AXES.Tx.url = "connect";
			AXES.Tx.command = "REOPEN";
			AXES.Tx.cursorRow = this.cursor.row;
			AXES.Tx.cursorCol = this.cursor.col;
			AXES.Tx.aidKey = "";
			AXES.Tx.data = data;
			AXES.Tx.sendReLogin();
		} else {
			AXES.Tx.sendHello();
		}
		this.busy = true;
		this.fire("ScreenPosted", [{ terminal: this }]);
		AXES.Layout.setBusy(true);
		return;
	}
	
	var inpFields = [];
	var modifiedInpFields = [];
	var includeData = true;
	var aidKey = this._aidKeyForPost;
	this._aidKeyForPost = null;
	
	// Check fKeyMask to see if the AID key should include screen data
	if ((aidKey > 0x30) && (aidKey < 0x3d)) {
		if (this.fKeyMask.charAt(aidKey - 0x31) == "0") includeData = false;
	}
	if ((aidKey > 0xb0) && (aidKey < 0xbd)) {
		if (this.fKeyMask.charAt(aidKey - 0xa5) == "0") includeData = false;
	}
	if (aidKey == AXES.Keyboard.aidKeys.Enter) {
		// Enter key. See if the field under the cursor is an autoselect field.
		var fldAtCursor = this.getFieldAtPosition(this.cursor);
		if (fldAtCursor && (fldAtCursor.type == "Select") && fldAtCursor.autoSelect) {
			fldAtCursor.selectOptionAtPosition(this.cursor);
		}
	}
	if ( includeData ) {
		var fld;
		var type;
		var fields = this.sortedFields;
		for ( var i in fields ) {
			fld = fields[ i ];
			type = fld.type;
			if ( "IField" == type || "Select" == type || "ScrollBar" == type ) {
				inpFields.push(fld);
				if ( fld.isModified() ) {
				    modifiedInpFields.push( fld );
				}
			}
		}
	}
	var fldsToSend = [];
	switch (this.postAction) {
		case 2: // read
			// Send all fields but only if at least one is modified
			AXES.Tx.command = "READ";
			fldsToSend = (modifiedInpFields.length > 0) ? inpFields : [];
			break;
		case 3: // read mdt
			AXES.Tx.command = "READ-MDT";
			fldsToSend = modifiedInpFields;
			break;
		case 4: // read mdt alt
			AXES.Tx.command = "READ-MDT-ALT";
			fldsToSend = modifiedInpFields;
			break;
		case 5: // read mdt alt
			AXES.Tx.command = "READ-IMM";
			fldsToSend = inpFields;
			break;
	}

	var dataString = "";
	for (var i = 0, fld; fld = fldsToSend[i]; i++) {
		if (fld.errorMessage != "") {
			var msg = TEXT.Core("Unable to post screen because some fields contain errors.")
			if (this.fire("ScreenPostFailed", [{ terminal: this, aidKey: aidKey, reason: msg }])) {
				AXES.Layout.onPostFailed(aidKey, msg);
			}
			return;
		}
		dataString += fld.getServerString();
	}
	
	AXES.Tx.cursorRow = this.cursor.row;
	AXES.Tx.cursorCol = this.cursor.col;
	AXES.Tx.aidKey = aidKey;		
	AXES.Tx.data = dataString;
	AXES.Tx.send();
	this.busy = true;
	this.fire("ScreenPosted", [{ terminal: this }]);
	AXES.Layout.setBusy(true);
	
	this.messageLight = false;

	for (var i = 0, fld; fld = this.fields[i]; i++) 
	{
		fld.screenUnloaded();
	}

	//
	this.isScreenRendered = false;
};
/**
 * (Rodel 5/03/2014) 
 * used by LongRange and RAMP applications 
 * - so their semantic/parameters or behaviours should not be modified without proper impact analysis. 
 */
AXES.Terminal.doSysPost = function(command, data, forceClose) {
	if (this.busy && !forceClose) 
	{
		if (!confirm(TEXT.core("DefaultMessage.AXS8820"))) return false;
	}
	AXES.Tx.command = command;
	AXES.Tx.cursorRow = this.cursor.row;
	AXES.Tx.cursorCol = this.cursor.col;
	AXES.Tx.aidKey = "";
	AXES.Tx.data = (data) ? data : "";
	AXES.Tx.send();
	this.busy = true;
	AXES.Layout.setBusy(true);
	return true;
};
/**
 * Sets the cursor position. This only sets the internal record of the cursor position.
 * It does not change the field with focus.
 * @param {Object} position
 * @param {Integer} position.row
 * @param {Integer} position.col
 */
AXES.Terminal.setCursorPosition = function(position) {
	AXES.Terminal.cursor.set(position);
};
/**
 * Returns the field, if any, at the specified position.Only fields in the top window are checked.
 * It does not change the field with focus.
 * @param {Object} position
 * @param {Integer} position.row
 * @param {Integer} position.col
 * (Rodel 5/03/2014) 
 * used by LongRange and RAMP applications 
 * - so their semantic/parameters or behaviours should not be modified without proper impact analysis.   
 */
AXES.Terminal.getFieldAtPosition = function(position) 
{
	var win = this.windows[this.windows.length-1];
	for (var i = 0, fld; fld = win.fields[i]; i++) 
	{
		var r = position.row - fld.row;
		var c = position.col - fld.col;
		if ((r >= 0) && (r < fld.height) && (c >= 0) && (c < fld.width)) 
		{ return fld; }
	}
	return null;
};
AXES.Terminal.getNamedFields = function() {
	var retVal = [];
	var nameHash = this.fields.indexByName
	for (name in this.fields.indexByName) {
		retVal.push(this.fields[nameHash[name]]);
	}
	return retVal;
};
AXES.Terminal.getFieldByName = function(name) {
	var index = this.fields.indexByName[name];
	return (index != undefined) ? this.fields[index] : null;
};
/**
 * Returns a field with the supplied ID.
 * @param {String} id The field ID
 */
AXES.Terminal.getFieldById = function(id) {
	var index = this.fields.indexById[id];
	return (index != undefined) ? this.fields[index] : null;
};
/**
 * Returns a field with the supplied adjusted ID.
 * If no window is supplied and multiple fields have the same adjusted ID, the field from the 
 * top-most window is returned.
 * @see AXES.Terminal.Field.Base#adjustedId
 * @param {String} adjustedID
 * @param {String|Integer} [winId]
 * @returns {AXES.Terminal.Field}
 */
AXES.Terminal.getFieldByAdjustedId = function(adjustedID, winId)
{
	if (typeof winId == "string") 
	{
		var match = winId.match(/\d+/);
		winId = (match) ? parseInt(match[0]) : 0;
	}
	if (this.windows.length == 1 || winId == 1000) 
	{
		// Special case. If there are no windows or the window number is 0 then
		// the adjustedId is the same as the id.
		return this.getFieldById(adjustedID);
	}
	var winsToTry = (winId == undefined) ? this.windows : [{ idNum: winId }];
	var match = adjustedID.match(/(ifld|text)(.*)/);
	if (match.length > 2) 
	{
		var fldIDNum = parseInt(match[2], 16);
		for (var i = winsToTry.length - 1; i >= 0; i--) 
		{
			var id = match[1] + (fldIDNum + ((winsToTry[i].idNum * 2) << 12)).toString(16);
			var index = this.fields.indexById[id];
			if (index != undefined) return this.fields[index];
		}
	}
	return null;
};
/**
 * Returns an array of the fields in the topmost window that are in a specified row of the screen.
 * @param {Number} rowToGet the row (in screen co-ordinates) to return.
 * @returns {AXES.Terminal.Field.Base[]} An array of fields.
 */
AXES.Terminal.getFieldsByRow = function(rowToGet) {
	if (!rowToGet) rowToGet = this.rows;
	// Only process the topmost window.
	var flds = this.windows[this.windows.length - 1].fields;
	var result = [];
	var l = flds.length;
	for (var i=0, fld; fld=flds[i]; i++) {
		if (fld.row == rowToGet) result.push(fld);
		if (fld.row > rowToGet) break;
	}
	return result;
};
/**
 * Converts a field ID from an input field to an output field and vice versa. If the supplied
 * string is not an input or output field id it is returned unchanged.
 * @param {String} theID
 * @returns {String}
 */
AXES.Terminal.swapInputOutputFieldId = function(theID) 
{
	return theID.replace(/(ifld|text)/g, function($1)
	{
		switch ($1)
		{
			case "ifld": return "text";
			case "text": return "ifld";
			default: return $1;
		}
	});
};
/**
 * Handles the processing of Object Commands
 * @namespace
 */
AXES.Terminal.ObjectCommands = {
	process: function(commandNode) {
		var self = this;
		$(commandNode).children().each(function(index, node) {
			switch (node.nodeName) {
				case "DEL":
					var flds = AXES.Terminal.fields, id=$(node).attr("id");
					var index = flds.indexById[id];
					if (typeof index == "undefined") {
						var wins = AXES.Terminal.windows;
						for (var i=1;i<wins.length;i++) {
							var win = wins[i];
							if (win.id == id) {
								// A window is being deleted. Delete all it's fields (if they haven't been deleted already)
								for (var j=0;j<win.fields.length;j++) {
									var index = flds.indexById[win.fields[j].id];
									if (typeof index != "undefined") flds[index] = null;
								}								
								wins.splice(i,1);
								break;
							}
						}
						if (i>=wins.length) {
							AXES.SysTrace.warn(TEXT.Core("Attempt to delete a field that doesn't exist: %s1", $(node).attr("id")));
						}
					} else {
						flds[index] = null;
					}
					break;
				case "MOD":
					var $node = $(node);
					var flds = AXES.Terminal.fields;
					var index = flds.indexById[$node.attr("id")];
					if (index != undefined) {
						var fld = flds[index];
						switch (fld.type)
						{
							case AXES.Terminal.Field.Types.SELECT:
								if ($node.attr("flags") != undefined) fld.flags = $node.attr("flags");
								if ($node.attr("ind") != undefined) fld.ind = $node.attr("ind");
								if ($node.attr("lmar") != undefined) fld.lmar = Boolean(parseInt($node.attr("lmar")));
								if ($node.attr("autopost") != undefined) fld.autoPost = Boolean(parseInt($node.attr("autopost")));
								if ($node.attr("autoselect") != undefined) fld.autoSelect = Boolean(parseInt($node.attr("autoselect")));
								if ($node.attr("multi") != undefined) fld.multi = Boolean(parseInt($node.attr("multi")));
								if ($node.attr("drop") != undefined) fld.drop = Boolean(parseInt($node.attr("drop")));
								if ($node.attr("childwidth") != undefined) fld.childWidth = parseInt($node.attr("childwidth"));
								if ($node.attr("childcols") != undefined) fld.childCols = parseInt($node.attr("childcols"));
								if ($node.attr("childpad") != undefined) fld.childPad = parseInt($node.attr("childpad"));
								if ($node.attr("scrollbar") != undefined) fld.scrollbar = Boolean(parseInt($node.attr("scrollbar")));
								if ($node.attr("items") != undefined) fld.scrollRows = parseInt($node.attr("items"));
								if ($node.attr("slider") != undefined) fld.scrollPosition = parseInt($node.attr("slider"));
								if ($node.attr("type") != undefined) fld.selectType = $node.attr("type");
								fld.processChildNodes($node);
								break;
							case AXES.Terminal.Field.Types.SCROLLBAR:
								if ($node.attr("items") != undefined) fld.scrollRows = parseInt($node.attr("items"));
								if ($node.attr("slider") != undefined) fld.scrollPosition = parseInt($node.attr("slider"));
								if ($node.attr("moveCursor") != undefined) fld.moveCursor = Boolean(parseInt($node.attr("moveCursor")));
								if ($node.attr("mdt") != undefined) fld.mdt = Boolean(parseInt($node.attr("mdt")));
								break;
							case AXES.Terminal.Field.Types.OUTPUT:
								var newFld = new AXES.Terminal.Field.TextField($node.attr("obj",fld.obj));
								// Axes TS1 processed the "keep" attribute to determine which attributes
								// to keep. It then reset all others. But the keep value is hard coded 
								// to |obj|id|r|c|win| in the server.
								// So, just create a new Field and copy the few parameters that need to be kept.
								newFld["row"] = fld["row"];
								newFld["col"] = newFld["column"] = fld["col"];
								newFld["win"] = fld["win"];
								newFld["winNum"] = fld["winNum"];
								newFld.type = fld.type;
								flds[index] = newFld;
								break;
							case AXES.Terminal.Field.Types.INPUT:
								var newFld = new AXES.Terminal.Field.InputField($node.attr("obj",fld.obj));
								// Axes TS1 processed the "keep" attribute to determine which attributes
								// to keep. It then reset all others. But the keep value is hard coded 
								// to |obj|id|r|c|win| in the server.
								// So, just create a new Field and copy the few parameters that need to be kept.
								newFld["row"] = fld["row"];
								newFld["col"] = newFld["column"] = fld["col"];
								newFld["win"] = fld["win"];
								newFld["winNum"] = fld["winNum"];
								newFld.type = fld.type;
								flds[index] = newFld;
								break;
						}
					}
					break;
				default:
					var msg = TEXT.Core("Unknown DDS command: %s1", $(node).nodeName);
					AXES.onError(2, msg);
					break;
			}
		});
	}
};
/**
 * Handles the processing of DDS Commands
 * @namespace
 */
AXES.Terminal.DDSCommands = {
	WTDFlag: {
		RESET_AID: 0x0100,
		CLR_MDT_INP: 0x0200,
		CLR_MDT_ALL: 0x0400,
		NULL_MDT_INP: 0x0800,
		NULL_ALL_INP: 0x1000,
		CURS_MOVE: 0x2000
	},
	process: function(commandNode) {
		var self = this;
		var term = AXES.Terminal;

		$(commandNode).children().each(function(index, node) {
			switch (node.nodeName) {
				case "CLEAR":
					self.clear(node);
					break;
				case "CLEAR_FMT":
			        self.clearFormat(node);
					break;
				case "WRITE":
					self.write(node);
					break;
				case "DSPF":
					var $node = $(node);
				 	if (term.dspf.name != $node.attr("name")) {
						// Reset
						term.dspf = {};
						term.recordFormats = [];
					}
					jQuery.extend(term.dspf, {
						name: $(node).attr("name"),
						library: $(node).attr("library"),
						type: $(node).attr("type")
					});
			        break;
				case "UIM":
					var $node = $(node);
					if ($node.attr("list") == "1") jQuery.extend(term.dspf, {list: true});
					if ($node.attr("menu") != "1") {
						//if guiRules then guiRules.noMenu = true
						// (TODO: come back to this when implimenting autoGUI)
					}
					break;
				case "READ":
					term.readFlags = parseInt($(node).attr("flags"));
					switch ($(node).attr("type")) {
						case "all":
							term.postAction = 2; // read
							break;
						case "mdt":
							term.postAction = 3; // read MDT
							break;
						case "alt":
							term.postAction = 4; // read MDT ALT
							break;
						case "imm":
							term.postAction = 5; // read IMMEDIATE
							// AutoPost = true;
							break;
						default:
							term.postAction = 3; // read MDT
							break;
						}
					break;
				case "ERROR":
					var $node = $(node);
					var  s = $node.attr("text");
					var msg = s.replace(/[ \f\n\r\t\v]*$/,"");  // strip off trailing spaces
					term.messages.push(msg);
					if ($node.attr("r")) {
						term.cursor.set({
							row: parseInt($node.attr("r")),
							col: parseInt($node.attr("c"))
						});
					}
					break;
				default:
					var msg = TEXT.Core("Unknown DDS command: %s1", node.nodeName);
					AXES.onError(2, msg);
					break;
			}
		});
	},
	reset: function(flags) {
		/* Called after a transaction is sent to the server to 
		 * reset the terminal ready for the next transaction.
		 */
		var clr_mdt_inp = (flags & 0x02); // Clear mdt on all input fields
		var clr_mdt_all = (flags & 0x04); // Clear mdt on all fields
		var null_mdt_inp = (flags & 0x08); // Reset data on input fields
		for (var i = 0, f; f = AXES.Terminal.fields[i]; i++) {
			f.ddsReset(clr_mdt_all || (clr_mdt_inp && (f.type == AXES.Terminal.Field.Types.INPUT)));
		}
	},
	clear: function(node) {
		var term = AXES.Terminal;
		var alt = $(node).attr("alt");
		if (alt && (alt == "1")) {
			term.rows = 27;
			term.columns = term.cols = 132;
			term.messageRow = 28;
		} else {
			term.rows = 24;
			term.columns = term.cols = 80;
			term.messageRow = 25;
		}
		term.fields.length = 0;
		term.fields.indexById = {};
		term.fields.indexByName = {};
		term.sortedFields.length = 0;
		term.sequencedFields.length = 0;
		term.recordFormats = [];
		term.windows = [new AXES.Terminal.Window()];
		term.htmlItems.length = 0;
		term.fKeyMask = "111111111111111111111111";
		term.cursorHome = {row: 1, col: 1};
		term.cursor.set(term.cursorHome);
		term.dspf = {};
		term.transactionLog.length = 0;
	},
	clearFormat: function(node) {
		// Convert all input fields to text fields
		for(var i=0,f; f=AXES.Terminal.fields[i]; i++) {
			if (f.type == "IField") {
				f.Type = "Text";
				f.obj = "txt";
			}
		}
	},
	write: function(writeNode) {
		// writeNode is the WRITE node from the transaction XML
		var term = AXES.Terminal;
		term.writeFlags = Number($(writeNode).attr("flags"));
		if (term.signatureModel != 1) 
		{
			term.recordFormats = [];
		}
		$(writeNode).children().each(function(index, node) {
			var $node = $(node);
			switch (node.nodeName) {
				case "SOH":
					var fkMask = $(node).attr("fkmask");
					if (fkMask) {
						term.fKeyMask = fkMask;
					} else {
						if (!AXES.Tx.delta) term.fKeyMask = "111111111111111111111111";
					}
					var msgRow = parseInt($node.attr("msgrow"));
					term.messageRow = ((msgRow == NaN) || (msgRow == 0) || (msgRow > term.rows)) ? term.rows : msgRow;
					break;
				case "HC":
					var r = parseInt( $node.attr( "r" ) );
					var c = parseInt( $node.attr( "c" ) );
					// adjust the home cursor if not updated
					if ( 1 == r && 1 == c )
					{
						var list = $( writeNode ).children();
						var len = list.length;
						for ( var i = 0; i < len; i++ )
						{
							if ( "IField" == list[ i ].nodeName )
							{
								r = parseInt( $( list[ i ] ).attr( "r" ) );
								c = parseInt( $( list[ i ] ).attr( "c" ) );
								break;
							}
						}
					}
					term.cursorHome.row = r;
					term.cursorHome.col = c;
					break;
				case "IC":
					term.cursor.row = term.cursorHome.row = parseInt($node.attr("r"));
					term.cursor.col = term.cursorHome.col = parseInt($node.attr("c"));
					break;
				case "MC":
					term.cursor.row = parseInt($node.attr("r"));
					term.cursor.col = parseInt($node.attr("c"));
					break;
				case "HTML":
					break;
				case "RCDFMT":
					var rfmt;
					switch ($node.attr("type"))
					{
						case AXES.Terminal.RecordFormat.Types.WINDOW:
							rfmt = new AXES.Terminal.RecordFormat.Window($node);
							break;
						case AXES.Terminal.RecordFormat.Types.SUBFILE:
							rfmt = new AXES.Terminal.RecordFormat.Subfile($node);
							break;
						default:
							rfmt = new AXES.Terminal.RecordFormat.Basic($node);
					}
					// If modifying the previous screen then the record format may already
					// exist. Look for it using a key of type.name (in some rare cases
					// two record formats of different types may have the same name
					var rfmtKey = rfmt.type + "." + rfmt.name;
					var i = (term.recordFormats[rfmtKey] !== undefined) ? term.recordFormats[rfmtKey] : term.recordFormats.length;
					term.recordFormats[rfmtKey] = i;
					// Some code references by name
					// TODO: Code that references by name may need to be updated to include the type.
					term.recordFormats[rfmt.name] = i;
					term.recordFormats[i] = rfmt;
					break;
				case "Text":
					var fld = new AXES.Terminal.Field.TextField($node);
					term.fields.push(fld);
					break;
				case "IField":
					var fld = new AXES.Terminal.Field.InputField($node);
					term.fields.push(fld);
					break;
				case "Select":
					var fld = new AXES.Terminal.Field.SelectField($node);
					term.fields.push(fld);
					break;
				case "ScrollBar":
					var fld = new AXES.Terminal.Field.ScrollbarField($node);
					term.fields.push(fld);
					break;
				case "Window":
					var win = new AXES.Terminal.Window($node);
					term.windows.push(win);
					break;
				case "HTMLKW":
					var fld = new AXES.Terminal.Field.HtmlField($node);
					term.fields.push(fld);
					break;
				default:
			}
		});
	}
};
/**
 * Represents a window in the 5250 terminal.
 * @namespace
 * @constructor
 * @param {Object} windowNode A jQuery object containing the XML node from the server that defines the window.
 * @property {Number} row The start row of the window.
 * @property {Number} col	The start column of the window.
 * @property {Number} width The width of the window.
 * @property {Number} height The height of the window.
 * @property {AXES.Terminal.Field.Base[]} fields An array of the fields in the window.
 */
AXES.Terminal.Window = function(windowNode) {
	this.isRealWindow = false;
	if (windowNode) 
	{
		if (windowNode instanceof AXES.Terminal.RecordFormat.Window) 
		{
			// A 5250 window has top/bottom borders of 1 character and 
			// Left/right borders of 3 characters (blank either side of 
			// the line). The position supplied in the transaction is the
			// position of the blank character to the left of the border.
			// The height and width supplied in the transaction are the
			// size of the inner area of the window.
			this.id = "winA";
			this.idNum = 1000;
			this.obj = "gui";
			this.row = windowNode.startLine;
			this.height = windowNode.endLine - this.row - 1;
			this.col = windowNode.startCol - 1;
			this.width = windowNode.endCol - this.col - 4;
			this.capCursor = true;
			this.pulldown = false;
		} else {
			this.id = windowNode.attr("id");
			this.idNum = parseInt(this.id.match(/\d+/)[0]);
			this.obj = windowNode.attr("obj");
			this.row = parseInt(windowNode.attr("r"));
			this.height = parseInt(windowNode.attr("rr"));
			this.col = parseInt(windowNode.attr("c"));
			this.width = parseInt(windowNode.attr("cc"));
			this.capCursor = (windowNode.attr("capcurs") == "1");
			this.pulldown = (windowNode.attr("pulldown") == "1");
			this.isRealWindow = true;
			var self = this;
			windowNode.children().each(function(index, node) {
				switch (node.nodeName) {
					case "WinTitle":
						var $node = $(node);
						var title = {
							align: $node.attr("align"),
							text: $node.attr("text"),
							attribute: $node.attr("attr")
						};
						if ($node.attr("footer")) {
							self.footer = title;
						} else {
							self.header = title;
						}
						break;
					default:
						var msg = TEXT.Core("Unknown Window Node: %s1", node.nodeName);
						AXES.onError(2, msg);
						break;
				}
			});
		}
	} else {
		this.id = "win0";
		this.idNum = 0;
		this.obj = "gui";
		this.row = 0;
		this.height = AXES.Terminal.rows;
		this.col = 0;
		this.width = AXES.Terminal.cols;
		this.capCursor = true;
	}
	this.fields = [];
};
/**
 * Returns an array of the fields in the window that are in a specified row of the screen.
 * @param {Number} rowToGet the row (in screen co-ordinates) to return.
 * @returns {AXES.Terminal.Field.Base[]} An array of fields.
 */
AXES.Terminal.Window.prototype.getFieldsByRow = function(rowToGet) {
	if (!rowToGet) rowToGet = this.rows;
	var sFld = this.fields;
	var result = [];
	for (var i = sFld.length-1, fld; fld=sFld[i]; i--) {
		if (fld.row == rowToGet) result.unshift(fld);
		if (fld.row < rowToGet) break;
	}
	return result;
}




/**
 * @namespace
 */
AXES.Terminal.RecordFormat = {
	Types: {
		RECORD: "record",
		SUBFILE: "sfl",
		SUBFILE_HEADER: "sflhdr",
		SUBFILE_CONTROL: "sflctl",
		SUBFILE_MESSAGE_RECORD: "sflmsgrcd",
		WINDOW: "window"
	}
};

AXES.Terminal.RecordFormat.Basic = AXES.EventsClass.extend(
	/** @lends AXES.Terminal.RecordFormat.Basic.prototype */
	{
	/** 
	 * Represents a RecordFormat in the 5250 screen.
	 * @constructs
	 * @augments AXES.EventsClass
	 * @param {RCDFMTNode} node A jQuery object containing the XML node from the server that defines the RecordFormat.
	 * @property {String} name The name of the Record Format
	 * @property {AXES.Terminal.RecordFormat.Types} type The Record Format type
	 * @property {number} startLine The first line of the Record Format
	 * @property {number} endLine The last line of the Record Format
	 * @property {AXES.Terminal.Field[]} fields An array of the fields in the Record Format
	 */
		init: function(RCDFMTNode)
		{
			this.name = RCDFMTNode.attr("name");
			this.type = RCDFMTNode.attr("type");
			this.startLine = parseInt(RCDFMTNode.attr("strrow"));
			this.endLine = parseInt(RCDFMTNode.attr("endrow"));
			this.fields = [];
		}
	});

AXES.Terminal.RecordFormat.Window = AXES.Terminal.RecordFormat.Basic.extend(
	/** @lends AXES.Terminal.RecordFormat.Window.prototype */
	{
	/** 
	 * Represents a RecordFormat in the 5250 screen.
	 * @constructs
	 * @augments AXES.Terminal.RecordFormat.Basic
	 * @param {RCDFMTNode} node A jQuery object containing the XML node from the server that defines the RecordFormat.
	 */
		init: function(RCDFMTNode)
		{
			this._super(RCDFMTNode);
			this.startCol = parseInt(RCDFMTNode.attr("wdwstrcol"));
			this.endCol = parseInt(RCDFMTNode.attr("wdwendcol"));
		}
	});

AXES.Terminal.RecordFormat.Subfile = AXES.Terminal.RecordFormat.Basic.extend(
	/** @lends AXES.Terminal.RecordFormat.Subfile.prototype */
	{
	/** 
	 * Represents a Subfile RecordFormat in the 5250 screen.
	 * @constructs
	 * @augments AXES.Terminal.RecordFormat.Basic
	 * @param {RCDFMTNode} node A jQuery object containing the XML node from the server that defines the RecordFormat.
	 * @property {number} rowsPerItem The number of screen rows per item (subfile row)
	 * @property {number} rowsInHeader The number of screen rows in the subfile header
	 * @property {AXES.Terminal.Field} moreIndicator The terminal field used to indicate more rows
	 * @property {number} maxItems The maximum number of items that can be displayed on the screen.
	 * @property {Object} cols A hash of {@link AXES.Terminal.RecordFormat.SubfileColumn} indexed by start position of the cell as an offset from the start of the entry.
	 */
		init: function(RCDFMTNode)
		{
			this._super(RCDFMTNode);
			this.rowsPerItem = 1;
			this.rowsInHeader = null;
			this.moreIndicator = null;
			this.itemCount = 0;
			this.maxItems = 0;
			this.cols = {};
		},
		calculateStructure: function(rfDefinition)
		{
			rfDefinition = rfDefinition || {};
			if (!rfDefinition.rowsPerItem) rfDefinition.rowsPerItem = 1;
			if (rfDefinition.rowsInHeader == null) 
			{
				// The number of lines in the header is everything before the subfile up to 
				// a blank line (with a maximum of 3). If a blank line is not found, default to 1
				rfDefinition.rowsInHeader = 1;
				var s = this.startLine - 1;
				var e = Math.max(this.startLine - 5, 0);
				for (var i = s; i > e; i--) 
				{
					if (AXES.Terminal.getFieldsByRow(i).length == 0) 
					{
						rfDefinition.rowsInHeader = s - i;
						break;
					}
				}
			}
			this.rowsInHeader = rfDefinition.rowsInHeader;
			var rpi = this.rowsPerItem = rfDefinition.rowsPerItem;
			
			this.maxItems = Math.floor((this.endLine - this.startLine + 1) / rpi);
			this.moreIndicator = null;
			this.cols = {};
			
			var col;
			var subfileRow = 0;
			var rightmostCol = 0;
			var rowItems = [];
			for (var i = 0, field; field = this.fields[i]; i++) 
			{
				if (field.byp) continue;
				var newSubfileRow = Math.floor((field.row - this.startLine) / rpi);
				if (newSubfileRow > subfileRow) subfileRow = newSubfileRow;
				var rowOffset = (field.row - this.startLine) % rpi;
				if (field.column > rightmostCol) rightmostCol = field.column;
				var key = ((rowOffset * AXES.Terminal.cols) + field.column).toString();
				
				col = this.cols[key];
				if (!col) 
				{
					col = this.cols[key] = new AXES.Terminal.RecordFormat.SubfileColumn(this.maxItems);
					if ((rfDefinition.colNames != undefined) && (rfDefinition.colNames[key] != undefined) && (rfDefinition.colNames[key] != "")) 
					{
						// Column name is normally defined by the element in the 
						// first row.  If execution gets to here then the first
						// row element is missing on this particular screen.
						// Assign the row to this element (the first non-empty row).
						field.symbolicName = rfDefinition.colNames[key];
						AXES.Terminal.fields.indexByName[field.symbolicName] = AXES.Terminal.fields.indexById[field.id];
					}
				}
				col.setItem(subfileRow, field);
				field.subfileCol = key;
				field.subfileRow = subfileRow + 1;
			}
			this.itemCount = subfileRow + 1;
			
			// col will be the last non-empty column on the last line. If it only has
			// one item and that one item is the last row and it is the rightmost, 
			// then it is probably the more indicator.
			if (col && col.count == 1) 
			{
				var field = col.getItem(this.maxItems-1);
				if (field && (field.row == this.endLine) && (field.col == rightmostCol)) this.moreIndicator = field;
			}
		},
		/**
		 * Returns an array of {@link AXES.Terminal.Field} objects in a subfile entry.
		 * @param {number} index The subfile entry to retrieve.
		 * @returns {AXES.Terminal.Field[]}
		 */
		getEntry: function(index)
		{
			var theEntry = null;
			if ((this.cols) && (index <= this.itemCount) && (index > 0)) 
			{
				if (this.sortedColIndexes == undefined) 
				{
					this.sortedColIndexes = [];
					for (var key in this.cols) 
					{
						this.sortedColIndexes.push(key);
					}
					this.sortedColIndexes.sort(function(a, b)
					{
						return Number(a) - Number(b);
					});
				}
				var idx = index - 1
				theEntry = [];
				for (var i = 0, key; key = this.sortedColIndexes[i]; i++) 
				{
					var el = this.cols[key].getItem(idx);
					if (el == undefined) el = null;
					theEntry.push(el);
				}
			}
			return theEntry;
		},
		/**
		 * Returns an array of {@link AXES.Terminal.Field} objects in a subfile entry.
		 * @param {fld} fld an entry where it belong to a subfile to retrieve.
		 * @returns {AXES.Terminal.Field[]}
		 */        
		getSubfileColumn: function(fld)
		{
		    var col = {};
		    var cols = fld.recordFormat.cols || [];
		    // check if field belong to subfile columns
		    for (var subfile in cols) {
		        for (var cfld in cols[subfile].items) {
		            if ( (cols[subfile].items[cfld]) && (fld.id == cols[subfile].items[cfld].id) ) {
		                col[subfile] = cols[subfile];
		                break;
		            }
		        }
		        if (col > 0) break;
		    }
		    return col;
		}
	});

/**
 * Represents a subfile column
 * @class
 * @param {number} maxItems The maximum number of items in the subfile
 */
AXES.Terminal.RecordFormat.SubfileColumn = function(maxItems) 
{
	this.items = new Array(this.maxItems);
	this.count = 0;
}
AXES.Terminal.RecordFormat.SubfileColumn.prototype.setItem = function(subfileRow, field) 
{
	if (this.items[subfileRow] === undefined) this.count++;
	if (field === null || field === undefined) this.count--;
	this.items[subfileRow] = field;
}
AXES.Terminal.RecordFormat.SubfileColumn.prototype.getItem = function(subfileRow) 
{
	return this.items[subfileRow];
}

/**
 * @namespace
 */
AXES.Terminal.Field = {

	/**
	 * Possible field types
	 * @namespace
	 */
	Types: {
		/** @constant */
		OUTPUT: "Text",
		/** @constant */
		INPUT: "IField",
		/** @constant */
		SELECT: "Select",
		/** @constant */
		SCROLLBAR: "ScrollBar",
		/** @constant */
		HTML: "HTMLKW"
	},
	/**
	 * Possible Ideographic Styles ({@link AXES.Terminal.Field.InputField.ideo}).
	 * @namespace
	 */
	IdeoStyles: {
		/** @constant */
		Unknown: -1, /* Ideo not set or unknown. */
		/** @constant */
		DBCSOnly: 1, /* Always DBCS. Field wrapped with shifts. */
		/** @constant */
		Data: 2, /* Always DBCS. Shifts are in extended attr not field. */
		/** @constant */
		Either: 3, /* DBCS strings are wrapped in shifts. */
		/** @constant */
		Open: 4 /* DBCS strings are wrapped in shifts. */
	}
}

/**
 * Indicates that the field value has changed.  If you wish to indicate
 * that the value is an error, set the errorMessage property (a HasError 
 * event will be raised).
 * @name AXES.Terminal.Field.Base#ValueChanged
 * @event
 * @param {EventObject} e Event parameters.
 * @param {AXES.Terminal.Field.Base} e.field The Field instance.
 */
/**
 * Indicates that the field value has an error.
 * @name AXES.Terminal.Field.Base#HasError
 * @event
 * @param {EventObject} e Event parameters.
 * @param {AXES.Terminal.Field.Base} e.field The Field instance.
 */
/**
 * Indicates that the setFocus method has been called requesting that the focus
 * be set to the field. An extension that creates user interface elements should
 * listen for this event and update the UI accordingly.
 * @name AXES.Terminal.Field.Base#SetFocus
 * @event
 * @param {EventObject} e Event parameters.
 * @param {AXES.Terminal.Field.Base} e.field The Field instance.
 */
/**
 * Indicates that the field should be highlighted. This is triggerred when the
 * field is selected in the developer tools.  An extension that creates user
 * interface elements may need to listen for this event and update the UI
 * accordingly.
 * @name AXES.Terminal.Field.Base#SetHighlight
 * @event
 * @param {EventObject} e Event parameters.
 * @param {AXES.Terminal.Field.Base} e.field The Field instance.
 */
/**
 * Indicates that the field has been added to to the Document Object Model.
 * Field extensions are initialised before being placed into the terminal.
 * An extension that needs to perform actions after it has been added to
 * the DOM should listen for this event and perform those actions in the
 * handler.
 * @name AXES.Terminal.Field.Base#AddedToDOM
 * @event
 * @param {EventObject} e Event parameters.
 * @param {AXES.Terminal.Field.Base} e.field The Field instance.
 */

AXES.Terminal.Field.Base = AXES.Extensions.ExtendableObject.extend(
	/** @lends AXES.Terminal.Field.Base.prototype */
	{
	/** 
	 * Represents a field in the 5250 screen. This is a base class used by other, more specific field types, and will not exist as itself.
	 * @constructs
	 * @augments AXES.Extensions.ExtendableObject
	 * @param {Object} node A jQuery object containing the XML node from the server that defines the field.
	 * @property {AXES.Terminal.Field.Types} type The type of the field.
	 * @property {String} errorMessage Description of any error in the field value.
	 */
		init: function(fieldNode)
		{
			this._super();
			if (fieldNode) 
			{
				this.type = fieldNode[0].nodeName;
				this.id = fieldNode.attr("id");
				this.obj = fieldNode.attr("obj");
				this.win = fieldNode.attr("win") || "";
				this.winNum = (this.win) ? parseInt(this.win.match(/\d+/)[0]) : 0;
				this.row = parseInt(fieldNode.attr("r"));
				this.col = this.column = parseInt(fieldNode.attr("c"));
				this.byp = Boolean(parseInt(fieldNode.attr("byp")));
				this.mdt = Boolean(parseInt(fieldNode.attr("mdt")));
			}
			this.recordFormat = null;
			this.data = "";
			this.initialData = "";
			this.errorMessage = "";
			this.isInput = false;
			this.symbolicName = "";
			this.addEvents(["ValueChanged", "SetFocus", "SetHighlight", "SetCaretPosition", "HasError", "AddedToDOM"]);
		},
		
		/**
		 * The ID encodes the window number which varies depending on the order
		 * windows are opened. This returns an ID without the window number component
		 * giving a more reliable value for use in pattern matching or when saving the 
		 * field ID for later use.
		 * @returns {String} The adjusted ID
		 */
		adjustedId: function()
		{
			if ((!this.win) || (this.win == "")) 
				return this.id;
			if (this._adjustedId) 
				return this._adjustedId;
			this._adjustedId = this.id;
			var match = this.win.match(/\d+/);
			if (match) {
				var winNum = (parseInt(match[0]) * 2) << 12;
				match = this.id.match(/(ifld|text|seln|html)(.*)/);
				if (match && match.length > 2) 
				{
					var idNum = parseInt(match[2], 16);
					idNum = idNum - winNum;
					this._adjustedId = match[1] + idNum.toString(16);
				}
			}
			return this._adjustedId;
		},
		
		/**
	 	* @private
	 	*/
		notifyValueListeners: function()
		{
			this.fire("ValueChanged", [{
				field: this
			}]);
			if (this.errorMessage != "") 
			{
				this.fire("HasError", [{
					field: this
				}]);
			}
		},
		/**
		 * Called when a screen is sent to the host to reset the field
		 * @param {Boolean} resetMdt
		 */
		ddsReset: function(resetMdt) {
			this.mdt = false;
		},
		/**
	 	* Informs the field that it has been added to the DOM.
	 	*/
		addedToDOM: function()
		{
			this.fire("AddedToDOM", [{
				field: this
			}]);
		},
		/**
	 * Returns the current value of the field.
	 * @returns {String}
	 */
		getValue: function()
		{
			return this.data;
		},
		/**
		 * Returns an array representing the fields footprint on the screen
		 */
		getTextArray: function() {
			var h = (typeof this.height == "number") ? this.height : 1;
			var w = (typeof this.width == "number") ? this.width : 0;
			var a = new Array(h);
			var s = this.data.split(/\n/);
			for (var i=0;i<h;i++) {
				a[i] = (s[i]) ? s[i].rpad(w) : "".rpad(w);
			}
			return a;
		},
		/**
	 * Sets the value of the field. If the value of the field is changed the
	 * ValueChanged event will be triggered.  If there is an error setting the value,
	 * false will be returned.
	 * @param {String} newVal
	 * @returns {Boolean}
	 */
		setValue: function(newVal)
		{
			if (typeof newVal == "number") newVal = newVal.toString();
			if (!AXES.Utils.isString(newVal)) 
			{
				return this._setValueError(this, TEXT.Core("Field value must be a String."));
			}
			if (this.data != newVal) 
			{
				this.data = newVal;
				this.notifyValueListeners();
			}
			this.errorMessage = "";
			return true;
		},
		/**
	 * 
	 * @private
	 * @param {AXES.Terminal.Field.InputField} fld
	 * @param {String} errMsg
	 */
	_setValueError: function(fld, errMsg)
	{
		fld.errorMessage = errMsg;
		fld.notifyValueListeners();
		AXES.Layout.displayMessage(errMsg, true, true);
		return false;
	},
		/**
		 * Resets the field value to the initial value sent by the server.
		 */
		resetValue: function() {
			fld.errorMessage = "";
			if (this.data != this.initialData) 
			{
				this.data = this.initialData;
				this.notifyValueListeners();
			}
		},
		/**
	 	* Returns true if the field value, or the value of any CTD linked field, has been modified from what
	 	* was sent by the host.
	 	*/
		isModified: function()
		{
			if (this._isModified()) return true;
			var f = this;
			while (f.ctd_from) {
				f = AXES.Terminal.getFieldById("ifld" + f.ctd_from);
				if (f._isModified()) return true;
			}
			f =	this;
			while (f.ctd_to) {
				f = AXES.Terminal.getFieldById("ifld" + f.ctd_to);
				if (f._isModified()) return true;
			}
			return false;
		},
		/**
		 * Returns true if the field value has been modified from what
		 * was sent by the host.
		 */
		_isModified: function()
		{
			return ((this.data != this.initialData) || this.mdt);
		},
		/**
	 	 * Returns true if the field has been determined to be part of a subfile
	 	 */
		isIndexable: function()
		{
			if (this.subfileCol) return true;
			return false;
		},
		/**
		 * If the field is indexable, returns the number of rows available. Otherwise returns 1.
		 */
		length: function()
		{
			return (this.subfileCol) ? this.recordFormat.itemCount : 1;
		},
		/**
		 * Returns the position of this field in the subfile, 1 if it is not indexable.
		 */
		getSubfileIndex: function()
		{
			return (this.subfileRow) ? this.subfileRow : 1;
		},
		/**
		 * If a field is indexable, returns the field at the specifies index. If there is no 
		 * field at the requested index null is returned. If the field is not indexable, 
		 * it will return itself.
		 * @param {Number} index
		 */
		getIndexedSubfileField: function(index)
		{
			if (index == undefined || isNaN(index)) return this;
			if (this.isIndexable()) 
			{
				var col = this.recordFormat.cols[this.subfileCol];
				var el = (col) ? col.getItem(index - 1) : null;
				if (el == undefined) el = null;
				return el;
			}
			else 
			{
				return this;
			}
		},
		/**
	 * Will attempt to set the focus to the field in the UI. This is not guaranteed.
	 * It will send a SetFocus event to any objects registered as listeners. It is
	 * up to those objects to listen for the event and handle it appropriately.
	 * @param {Integer} [offset] The position of the cursor within the field.  This is not guaranteed. Support of the offset value is up to individual extension implimentations.
	 */
		setFocus: function(offset)
		{
			this.setCursor(offset);
			this.fire("SetFocus", [{
				field: this,
				offset: offset
			}]);
		},
		/**
	 * Will attempt to highlight the field in the UI. This is not gauranteed.
	 * It will send a SetHighlight event to any objects registered as listeners. It is
	 * up to those objects to listen for the event and handle it appropriately.
	 * @param {Boolean} highlight
	 */
		setHighlight: function(highlight)
		{
			this.fire("SetHighlight", [{
				field: this,
				highlight: highlight
			}]);
		},
		/**
	 * Sets the terminal cursor position to the field's position.  This updates
	 * the internal representation of the screen and the cursor position that is
	 * sent to the host.  It does not affect the user interface in any way.  If you
	 * want to move the cursor in the user interface, use setFocus instead.
	 * @param {Integer} [offset=0] The position of the cursor within the field.
	 */
		setCursor: function(offset)
		{
			if (isNaN(offset)) 
			{
				// If offset is not specified and the cursor is already in the field, leave it alone.
				// Otherwise, set it to zero.
				
				if (AXES.Terminal.cursor.isInField(this)) return;
				offset = 0;
			}
			offset = parseInt(offset);
			var newPos = {
				col: this.col,
				row: this.row
			}
			if (offset > 0) 
			{
				newPos.col += offset % this.width;
				newPos.row += Math.floor(offset / this.width);
			}
			if (newPos.row >= (this.row + this.height)) 
			{
				newPos.col = this.col + this.width - 1;
				newPos.row = this.rol + this.height - 1;
			}
			AXES.Terminal.cursor.set(newPos);
		},
		getHTMLElement: function()
		{
			return this.terminalDiv;
		},
		setVisible: function(vis, index)
		{
			var el = this.getHTMLElement(index);
			if (vis) 
			{
				$(el).css("visibility", "visible");
			}
			else 
			{
				$(el).css("visibility", "hidden");
			}
		},
		/**
		 * Returns the size of the field converted into pixels. This is a computed value based
		 * on the current terminal character size and can be used for setting the size of
		 * elements. To determine the actual size of elements once added to the DOM, use getSize.
		 * @returns {Object} size
		 * @returns {Integer} size.width
		 * @returns {Integer} size.height
		 */
		getComputedSize: function()
		{
			// Since the field dimensions are defined in characters, a better name for this might be something
			// like getSizeInPixels - but, to ensure compatability with existing eXtensions we need to use
			// getSize.	
			return AXES.Layout.characterDimensionsToPixels({
				width: this.width ? this.width : 0,
				height: this.height ? this.height : 0
			}, false);
		},
		/**
		 * Returns the actual size of the DIV containing this field in the terminal as defined
		 * in the screen design, otherwise returns the computed values.
		 * @returns {Integer} width
		 * @returns {Integer} height
		 */
		getSize: function()
		{
			var compSize = this.getComputedSize();
			var retVal = {width: 0,height: 0};
			var $div = $(this.terminalDiv);
			if ($div.length) 
			{
				var s = $div[0].style;
				retVal.width = (s.width == "") ? compSize.width : $div.width();
				retVal.height = (s.height == "") ? compSize.height : $div.height();
			}
			return retVal;
		},
		isHidden: function()
		{
			return (this.attribute && this.attribute.match(/x(27|2F|37|3F)/gi));
		},
		/**
		 * Returns a string representation of this Field for use in logs
		 */
		toDebugString: function()
		{
			var txt = "5250 field, " + ((this.symbolicName) ? ("name = " + this.symbolicName) : "name unassigned");
			txt += this.isIndexable() ? ("[" + this.getSubfileIndex() + "]") : "";
			txt += ", value = \"" + this.getValue() + "\"";
			return txt;
		},
		/**
		 * Tells the field that the screen has been posted. The field will clear all
		 * event listeners and take any other neccesary action to effectively disable
		 * itself until the next screen arrives.
		 */
		screenUnloaded: function()
		{
			this.removeAllListeners();
		}
	});
AXES.Terminal.Field.TextField = AXES.Terminal.Field.Base.extend(
/**
* @lends AXES.Terminal.Field.TextField.prototype
*/
{
	/** 
	 * Represents an output field in the 5250 screen
	 * @constructs
	 * @augments AXES.Terminal.Field.Base
	 * @param {Object} node A jQuery object containing the XML node from the server that defines the field.
	 */
	init: function(fieldNode)
	{
		this._super(fieldNode);
		if (fieldNode) {
			this.rr = this.height = parseInt(fieldNode.attr("rr")) || 1;
			this.cc = this.width = parseInt(fieldNode.attr("cc"));
			this.attribute = fieldNode.attr("attr");
			this.data = this.initialData = fieldNode.attr("data");
			this.byp = (this.data == "" && this.cc == 0);
		}
	}
});

AXES.Terminal.Field.InputField = AXES.Terminal.Field.Base.extend(
/**
* @lends AXES.Terminal.Field.InputField.prototype
*/
{
	/** 
	 * Represents an input field in the 5250 screen
	 * @constructs
	 * @augments AXES.Terminal.Field.Base
	 * @param {Object} node A jQuery object containing the XML node from the server that defines the field.
	 * @property {String} imeMode The value that UI elements representing this field should use for their <a href="https://developer.mozilla.org/en/CSS/ime-mode">ime-mode</a> CSS property
	 * @property {AXES.Terminal.Field.IdeoStyles} ideo The ideographic style (only applies to input fields).
	 */
	init: function(fieldNode)
	{
		var val, ffw, bits567, reStr, re;
		this._super(fieldNode);
		this.isInput = true;
		if (fieldNode) 
		{
			this.rr = this.height = parseInt(fieldNode.attr("rr")) || 1;
			this.cc = this.width = parseInt(fieldNode.attr("cc"));
			this.attribute = fieldNode.attr("attr");
			val = this.data = this.initialData = fieldNode.attr("data");
			this.len = parseInt(fieldNode.attr("len"));
			
			// Details of the Field Format Word and Field Control Word values can
			// be found here:
			// http://publib.boulder.ibm.com/infocenter/iseries/v6r1m0/index.jsp?topic=/apis/QsnSetFld.htm
			ffw = this.fieldFormatWord = parseInt(fieldNode.attr("FFW"));
			bits567 = (ffw & 0x700) >> 8;
			this.me = parseInt(fieldNode.attr("me"));
			this.isAutoEnter = ffw & 0x80; // Bit 8
			this.isFieldExitRequired = ffw & 0x40; // Bit 9
			this.isMonocase = ffw & 0x20; // Bit 10
			this.isFeatureInput = (bits567 == 6); // Bit 5-7 = 110
			this.isNumeric = (bits567 == 2 || bits567 == 3 || bits567 == 5 || bits567 == 7);
			this.rightAlign = (((ffw & 0x7) == 5) || ((ffw & 0x7) == 6) || (bits567 == 7) || (bits567 == 3)) // Bits 13-15 = 101 or 110 or signed numeric
			this.data = this.initialData = (this.isNumeric) ? val.trim() : ((this.rightAlign) ? val.ltrim() : val.rtrim());
			
			// Field Control Word
			this.ctd = parseInt(fieldNode.attr("ctd"));
			if (this.ctd) 
			{
				this.ctd_from = fieldNode.attr("ctd_from");
				this.ctd_to = fieldNode.attr("ctd_to");
			}
			this.chk = parseInt(fieldNode.attr("chk"));
			this.wrap = parseInt(fieldNode.attr("wrap"));
			this.trig = parseInt(fieldNode.attr("trig"));
			this.transp = parseInt(fieldNode.attr("transp"));
			this.ideo = parseInt(fieldNode.attr("ideo"));
			this.focusAttribute = fieldNode.attr("high");
			this.next_fld = fieldNode.attr("next_fld");
			this.tabNext = fieldNode.attr("next_curs");
			this.sel_aid = fieldNode.attr("sel_aid");
			
			if (this.ideo == AXES.Terminal.Field.IdeoStyles.DBCSOnly ||
			this.ideo == AXES.Terminal.Field.IdeoStyles.Data ||
			((ffw&0xE00)==0x400) /* Bit 11 - Katans Shift only CCS#153292*/) 
			{
				this.imeMode = "active";
			}
			else if (!this.ideo) 
			{
				// No ideo, deactivate IME
				this.imeMode = "inactive";
			}
			else 
			{
				this.imemode = "auto";
			}
			
			if ((this.height > 1) && this.wrap) 
			{
				// The host will pre-wrap the data by inserting spaces to
				// pad each line to the field width.  It's not as simple as
				// removing multiple spaces as these are allowed within a line.
				// so, break it into lines and rtrim each.
				reStr = "(.{1," + this.width + "})";
				re = new RegExp(reStr, "gi");
				this.data = this.data.replace(re, function($0)
				{
					return $0.rtrim() + " ";
				});
				this.data = this.data.rtrim();
			}
		}
	},
	
	ddsReset: function(resetMdt) {
		this.data = this.initialData = "";
		this._super(resetMdt)
	},
	
	/**
	 * Returns the field value formatted for sending to the host
	 * @returns {String}
	 * @private
	 */
	getServerString: function()
	{
		var formattedValue = this.data;
		var formatChar = ""; // Format character
		switch ((this.fieldFormatWord & 0x700) >> 8)
		{ // Bits 5-7
			case 3: //numeric only
				formatChar = "+";
				break;
			case 5: //digits only
				formatChar = "D";
				break;
			case 7: //signed numeric
				formattedValue = formattedValue.replace(/ /g,'');
				// Correct format is padded left or right
				// with a "-" or " " on the end.
				
				// Correct padding is unclear. The 1.3.x code left pads if there is
				// already a space at the beginning or if the value is negative
				// (it's not clear if this is a bug).  Otherwise it right-pads.
				// For the moment, we'll left pad everything.
				var re = /-/g;
				var isNegative = (re.test(formattedValue));
				formattedValue = formattedValue.replace(re, "").lpad(this.len - 1);
				formattedValue += (isNegative) ? "-" : " ";
				formatChar = "-";
				break;
			default:
				if (this.wrap) 
				{
					// Replace any line breaks with a space
					formattedValue = formattedValue.replace(/(\r\n|\n)/, " ");
					// Instead of a line break character, wrapping is done by 
					// padding each line out to the full width with spaces.
					var wrapAt = this.width;
					var reStr = "(?:(.{1," + (wrapAt - 1) + "})($|\\s))|(.{1," + wrapAt + "})";
					var re = new RegExp(reStr, "gi");
					formattedValue = formattedValue.replace(re, function($0, $1)
					{
						return $1.rpad(wrapAt);
					});
					formattedValue = formattedValue.substring(0, this.length);
				}
				formatChar = "A";
		}
		
		// TODO: DBCS handling. This may be the place to insert SI/SO characters into
		// DBCS strings.  Axes 1.3 didn't do this before sending the data but
		// it may be doing it on keypress (it's currently a bit unclear)
		
		formattedValue = escape(formattedValue);
		return [this.row, this.col, (this.isModified() ? "1" : "0"), formattedValue.length + 1, formatChar + formattedValue, ""].join("_");
		
	},
	/**
	 * Formats the supplied value as if the FieldExit, FieldPlus or FieldMinus key has been pressed.
	 * Unlike a real 5250, this does not clear the characters to the right of the cursor (the cursor
	 * position is unknown and not considered).
	 * @param {Object} value The value to be formatted
	 * @param {Object} key The key to format for ("FieldExit", "FieldMinus" or "FieldPlus")
	 * @returns The formatted value or undefined (if the key is not allowed for this field).
	 */
	doFieldKeyFormat: function(value, key) 
	{
		var retVal, padChar = "", fldShift = ((this.fieldFormatWord & 0x700) >> 8); // Bits 5-7
		
		switch (this.fieldFormatWord & 0x7)
		{ // Bits 13-15
			case 5: //right adj, 0 fill
				padChar = "0";
				break;
			case 6: //right adj, blank fill
				padChar = " ";
				break;
		}
		
		switch (fldShift)
		{ // Bits 5-7
			case 7: //signed numeric
				padChar = " ";
			case 3: //numeric only
				retVal = value.replace(/[^0-9.,]/g, "") + (key == "FieldMinus" ? "-" : " ");
				break;
			default:
				if (key == "FieldMinus") return undefined;
				retVal = value;
		}
		
		if (padChar != "") 
		{
			retVal = retVal.lpad(this.len, padChar);
		}
		return retVal;
	},
	/**
	 * Checks if the supplied string contains any characters that are not valid for this field.
	 * See here for more info:
	 * http://publib.boulder.ibm.com/infocenter/iseries/v6r1m0/index.jsp?topic=/apis/QsnSetFld.htm
	 * @param {String} theChars
	 * @param {Boolean} [allowMatchedInvalid=false] If true, invalid characters will be allowed if they already exist, in the same position, in the initial field value.
	 * @returns {Boolean}
	 */
	stringHasInvalidChars: function(theChars, allowMatchedInvalid)
	{
		allowMatchedInvalid = Boolean(allowMatchedInvalid);
		var re;
		switch ((this.fieldFormatWord & 0x700) >> 8)
		{
			case 0: // Alpha shift
			case 2: // Numeric shift
			case 4: // Katakana (cursor r->l)
			case 6: // feature input
				return false;
			case 1: // Alpha only
				var ac = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz,.\\- " +
				"\xc0\xc1\xc2\xc2\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf;" +
				"\xe0\xe1\xe2\xe2\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff;";
				re = new RegExp("[^" + ac + "]", "g");
				break;
			case 3: // Numeric only
				re = new RegExp("[^0123456789+,. -]", "g");
				break;
			case 5: // Digits only
				re = new RegExp("[^0123456789 ]", "g");
				break;
			case 7: // Signed Numeric
				re = new RegExp("[^0123456789 -]", "g");
				break;
			default:
				return false;
		}
		var match = re.exec(theChars);
		if (match && !allowMatchedInvalid) return true;
		var pos1, pos2 = -1, offset = null;
		while (match) 
		{
			pos1 = match.index;
			pos2 = this.initialData.indexOf(match[0],pos2+1);
			if (pos2 < 0) return true;
			if (offset === null) offset = pos2 - pos1;
			if (pos1 + offset != pos2) return true;
			match = re.exec(theChars);
		}
		return false;
	},
	
	/**
	 * Sets the value of the field. If the value of the field is changed the
	 * ValueChanged event will be triggered.  If there is an error in the value,
	 * false will be returned and the errorMessage property will contain a
	 * description of the error.
	 * @param {String} newVal
	 * @param {Boolean} ignoreInvalidCharacters - optional, defaults to false. If true axTerm will not validate the value and assume that whatever is writing it has checked it 
	 * @returns {Boolean}
	 */
	setValue: function(newVal, ignoreInvalidCharacters)
	{
		if ((ignoreInvalidCharacters == null) || (ignoreInvalidCharacters != true)) ignoreInvalidCharacters = false; 
 
		if (typeof newVal == "number") newVal = newVal.toString();
		if (!AXES.Utils.isString(newVal)) 
		{
			return this._setValueError(this, TEXT.Core("Field value must be a String."));
		}
				
		this.data = newVal;
		this.errorMessage = "";
		
		if (this.isMonocase) 
		{
			newVal = AXES.Utils.monocaseString(newVal);
		}

		if (!ignoreInvalidCharacters)
		{		
			if (this.stringHasInvalidChars(newVal, true)) 
			{
				return this._setValueError(this, TEXT.Core("Field contains invalid characters."));
			}
		}
		
		var modValue = null;
		if (this.chk == 1) modValue = 10;
		if (this.chk == 2) modValue = 11;
		if (modValue) 
		{
			if (!AXES.Utils.checkMod(newVal, modValue)) 
			{
				return this._setValueError(this, TEXT.Core("Invalid mod (%s1) value: %s2", modValue, newVal));
			}
		}
		
		switch ((this.fieldFormatWord & 0x700) >> 8)
		{ // Bits 5-7
			case 7: //signed numeric
				// Correct format is padded left or right
				// with a "-" or " " on the end.
				var negFound = false, m, re = /-/g;
				while (m = re.exec(newVal)) {
					if (negFound || (m.index > 0 && m.index < (newVal.length-1))) 
					{
						return this._setValueError(this, TEXT.Core("Invalid sign character (-)."));
					}
					negFound = true;
				}
				break;
		}
		
		var dbcsInfo = this._checkDBCS(newVal, true);
		if (dbcsInfo.error) 
		{
			return this._setValueError(this, dbcsInfo.error);
		}
		
		if (dbcsInfo.excess) 
		{
			// The string is too long.
			return this._setValueError(this, TEXT.Core("Too many characters entered. Must be no more than (%s1).", this.len));
		}
		
		if ((this.fieldFormatWord & 0x7) == 7) 
		{
			// Bits 13-15 == 111 - mandatory fill
			if (this.len > dbcsInfo.length) 
			{
				return this._setValueError(this, TEXT.Core("Field must be filled."));
			}
		}
		
		this.data = newVal;
		this.notifyValueListeners();
		
		return true;
	},
	
	/**
	 * Checks the length of the supplied string.  String lengths in JavaScript
	 * are based on the number of characters. On the i they are based on the
	 * number of bytes used.  This function will return the length of the
	 * string in bytes and, if necessary, will return a value truncated at
	 * the field's maximum length.
	 * @param {String} valToCheck The value to check
	 * @returns {Object.val} The supplied string. If the string is too long for the field, this will be truncated to a length that will fit.
	 * @returns {Object.excess} If the val property has been truncated, this will contain the string that was removed.
	 * @returns {Object.length} The length of val as the host sees it.
	 */
	checkLength: function(valToCheck)
	{
		return this._checkDBCS(valToCheck, false);
	},
	/**
	 * Checks the supplied string for DBCS validity.  Some fields can accept
	 * only SBCS characters, some only DBCS, some a mixture and some must be all
	 * of one type or another.  This method checks the supplied string
	 * according to the rules for this field. This method will also check
	 * the length of the string.  String lengths in JavaScript are based
	 * on the number of characters. On the i they are based on the number
	 * of bytes used.  This function will return the length of the string in
	 * bytes and, if necessary, will return a value truncated at the field's
	 * maximum length.
	 * @private
	 * @param {String} valToCheck The value to check
	 * @param {Boolean} checkChars If true, the string is checked for DBCS/SBCS correctness. If false, just the length is checked.
	 * @returns {Object.error} An error message indicating the type of DBCS/SBCS error present. Empty string if no error. Undefined if checkChars is false.
	 * @returns {Object.val} The supplied string. If the string is too long for the field, this will be truncated to a length that will fit.
	 * @returns {Object.excess} If the val property has been truncated, this will contain the string that was removed.
	 * @returns {Object.length} The length of val as the host sees it.
	 */
	_checkDBCS: function(valToCheck, checkChars)
	{
		var retVal = {
			val: valToCheck
		};
		
		// DBCS Handling.  check for DBCS or SBCS characters where they 
		// shouldn't be and calculate the byte length of the string
		// on the host (as this is what should be used for length
		// checks rather than the number of characters).
		
		var IdeoStyles = AXES.Terminal.Field.IdeoStyles;
		var errorMessageFunc;
		if (checkChars) 
		{
			var ideo = (this.ideo) ? this.ideo : IdeoStyles.Unknown;
			switch (ideo)
			{
				case IdeoStyles.DBCSOnly:
					/** @ignore */errorMessageFunc = function(bDBCS){
						return (bDBCS) ? "" : TEXT.Core("DBCS-Only field contains SBCS characters.");
					};
					break;
				case IdeoStyles.Data:
					/** @ignore */errorMessageFunc = function(bDBCS){
						return (bDBCS) ? "" : TEXT.Core("DBCS-Graphic field contains SBCS characters.");
					};
					break;
				case IdeoStyles.Either:
					/** @ignore */errorMessageFunc = function(bDBCS, bStartCharDBCS){
						return (bDBCS != bStartCharDBCS) ? TEXT.Core("DBCS-Either field contains both SBCS and DBCS characters.") : "";
					};
					break;
				case IdeoStyles.Unknown:
					/** @ignore */errorMessageFunc = function(bDBCS, bStartCharDBCS){
						return (bDBCS) ? TEXT.Core("SBCS field contains DBCS characters.") : "";
					};
					break;
				default:
					/** @ignore */errorMessageFunc = function(bDBCS, bStartCharDBCS)
					{
						return "";
					};
			}
		}
		
		var l = valToCheck.length;
		var bStartCharDBCS = AXES.Utils.isDBCS(valToCheck.charCodeAt(0));
		var hostLength = 0;
		var inDBCS = false;
		for (var i = 0; i < l; i++) 
		{
			var charCode = valToCheck.charCodeAt(i);
			var newHostLength = hostLength + 1;
			var bDBCS = AXES.Utils.isDBCS(charCode);
			if (bDBCS) 
			{
				newHostLength++;
				if (!inDBCS) 
				{
					if (this.ideo != IdeoStyles.Data) 
						newHostLength += 2; // Allow for SI/SO characters
					inDBCS = true;
				}
			}
			else 
			{
				inDBCS = false;
			}
			
			if (checkChars) 
			{
				retVal.error = errorMessageFunc(bDBCS, bStartCharDBCS);
				if (retVal.error) return retVal;
			}
			
			if (newHostLength > this.len) 
			{
				// Too long. Truncate.
				retVal.excess = retVal.val.substring(i);
				retVal.val = retVal.val.substring(0, i);
				break;
			}
			hostLength = newHostLength;
		}
		
		retVal.length = hostLength;
		return retVal;
	},
	
	setError: function(errorText)
	{
		this.errorMessage = errorText;
	},
	/**
	 * If the field is part of a CTD series, returns the first field in that series.
	 * Otherwise returns itself.
	 */
	getFirstInCTDSeries: function()
	{
		if (!this.ctd || this.ctd == 1) return this;
		var tmp = this;
		while (tmp.ctd_from) 
		{
			tmp = AXES.Terminal.getFieldById("ifld" + tmp.ctd_from);
		}
		return tmp;
	}
});

AXES.Terminal.Field.ScrollbarField = AXES.Terminal.Field.TextField.extend(
/**
* @lends AXES.Terminal.Field.ScrollbarField.prototype
*/
{
	/**
	 * Represents a scrollbar field in the 5250 screen
	 * @constructs
	 * @augments AXES.Terminal.Field.TextField
	 * @param {Object} node A jQuery object containing the XML node from the server that defines the field.
	 */
	init: function(fieldNode)
	{
		this._super(fieldNode);
		if (fieldNode) 
		{
			this.flags = fieldNode.attr("flags");
			this.scrollRows = parseInt(fieldNode.attr("items"));
			this.scrollPosition = parseInt(fieldNode.attr("slider"));
			this.scrollVertical = Boolean(parseInt(fieldNode.attr("vert")));
			this.moveCursor = Boolean(parseInt(fieldNode.attr("moveCursor")));
			this.data = this.initialData = "";
		}
	},
	ddsReset: function(resetMdt) {
		delete this.scrollAmount;
		this._super(resetMdt);
	},
	getServerString: function()
	{
		var formattedValue = "S" + ((this.scrollAmount) ? this.scrollAmount : "0");
		
		return [this.row, this.col, (this.isModified() ? "1" : "0"), formattedValue.length, formattedValue, ""].join("_");
	},
	setValue: function(newVal)
	{
		AXES.SysTrace.error(TEXT.Dev("Attempt to call %s1 on a Scrollbar field ignored.", "setValue"));
	},
	getValue: function()
	{
		AXES.SysTrace.error(TEXT.Dev("Attempt to call %s1 on a Scrollbar field ignored.", "getValue"));
		return null;
	},
	isModified: function()
	{
		return ((this.scrollAmount != undefined) || this.mdt)
	}
});

AXES.Terminal.Field.SelectField = AXES.Terminal.Field.ScrollbarField.extend(
/**
* @lends AXES.Terminal.Field.SelectField.prototype
*/
{
	/**
 	* Represents a select field in the 5250 screen
 	* @constructs
 	* @augments AXES.Terminal.Field.ScrollbarField
 	* @param {Object} node A jQuery object containing the XML node from the server that defines the field.
	 */
	init: function(fieldNode)
	{
		this._super(fieldNode);
		if (fieldNode) 
		{
			this.ind = fieldNode.attr("ind");
			this.lmar = Boolean(parseInt(fieldNode.attr("lmar")));
			this.autoPost = Boolean(parseInt(fieldNode.attr("autopost")));
			this.autoSelect = Boolean(parseInt(fieldNode.attr("autoselect")));
			this.multi = Boolean(parseInt(fieldNode.attr("multi")));
			this.drop = Boolean(parseInt(fieldNode.attr("drop")));
			this.childWidth = parseInt(fieldNode.attr("childwidth"));
			this.childCols = parseInt(fieldNode.attr("childcols"));
			this.childPad = parseInt(fieldNode.attr("childpad"));
			this.scrollbar = Boolean(parseInt(fieldNode.attr("scrollbar")));
			this.scrollVertical = true;
			this.selectType = fieldNode.attr("type");
			this.attribute = "";
			
			this.processChildNodes(fieldNode);
			this.data = this.initialData = this.getServerString();
		}
	},
	
	/** @private */
	processChildNodes: function(fieldNode) {
		this.selectItems = [];
		this.seperators = [];
		
		var self = this;
		fieldNode.children().each(function(index, node) {
			switch (node.nodeName) {
				case "SelItem":
					var $node = $(node);
					var selItem = {
						sequence: parseInt($node.attr("seq")),
						row: parseInt($node.attr("r")),
						col: parseInt($node.attr("c")),
						width: parseInt($node.attr("cc")),
						flags: $node.attr("flags"),
						aidKey: parseInt($node.attr("aid")),
						byp: Boolean(parseInt($node.attr("byp"))),
						selected: Boolean(parseInt($node.attr("dflt"))),
						text: $node.attr("text")
					};
					if (selItem.selected) self.data = selItem.text;
					self.selectItems.push(selItem);
					break;
				case "present":
					// This is ignored.
					break;
				case "MenuSep":
					var $node = $(node);
					var sep = {
						row: parseInt($node.attr("r")),
						style: $node.attr("style")
					}
					break;
				default:
					var msg = TEXT.Core("Unknown Select node: %s1", node.nodeName);
					AXES.onError(2, msg);
					break;
			}
		});
	},
	
	ddsReset: function(resetMdt)
	{
		if (this.selectItems.length > 0) 
		{
			for (var i = 0, sel; sel = this.selectItems[i]; i++) 
			{
				sel.selected = false;
			}
			this.selectItems[0].selected = true;
		}
		this.data = this.initialData = this.getServerString();
		this._super(resetMdt);
	},
	isItemSelected: function(item)
	{
		if (item < this.selectItems.length) 
		{
			return this.selectItems[item].selected;
		}
		return false;
	},
	setItemSelected: function(item, value)
	{
		if (item < this.selectItems.length) 
		{
			if (!this.multi && value) 
			{
				// make sure any other selected items are deselected.
				for (var i = 0, sel; sel = this.selectItems[i]; i++) 
				{
					sel.selected = false;
				}
			}
			this.selectItems[item].selected = value;
			this.data = this.getServerString();
			this.notifyValueListeners();
		}
	},
	toggleItemSelected: function(item)
	{
		if (item < this.selectItems.length) 
		{
			var value = !this.selectItems[item].selected;
			this.setItemSelected(item, value);
			return value;
		}
		return null;
	},
	selectOptionAtPosition: function(position) {
		var r = position.row - this.row;
		var c = position.col - this.col;
		for (var i = 0, sel; sel = this.selectItems[i]; i++) 
		{
			if ((sel.row == r) && (sel.col <= c) && ((sel.col + sel.width) > c)) 
			{
				this.setItemSelected(i, true);
			}
		}
	},
	getServerString: function()
	{
		var formattedValue = "";
		if (this.multi) 
		{
			formattedValue = "X";
			for (var i=0,sel; sel=this.selectItems[i]; i++) {
				formattedValue += (sel.selected) ? "Y" : "N";
			}
			// The Select element represents a grid of options. Not all spaces
			// in the grid necessarily contain an option but we need to return
			// a value for every space
			formattedValue += new Array((this.height * this.childCols) - this.selectItems.length + 1).join("N");
		}
		else 
		{
			switch (this.ind)
			{
				case "btn":
					formattedValue += "Z";
					break;
				default:
					formattedValue += "Y";
			}
			var selNum = 0;
			for (var i=0,sel; sel=this.selectItems[i]; i++) {
				if (sel.selected) {
					selNum = sel.sequence;
					break;
				}
			}
			formattedValue += selNum;
		}
		
		if (this.scrollbar) 
		{
			formattedValue += "S" + ((this.scrollAmount) ? this.scrollAmount : "0");
		}
		
		return [this.row,
	   			this.col,
				(this.isModified() ? "1" : "0"),
				formattedValue.length,
				formattedValue, "" ].join("_");
	},
	/**
	 * Does nothing. Use setItemSelected, toggleItemSelected or selectOptionAtPosition instead
	 */
	setValue: function()
	{
		AXES.SysTrace.error(TEXT.Dev("Attempt to call %s1 on a Select field ignored.", "setValue"));
	},
	/**
	 * Does nothing. Use isItemSelected instead.
	 * @returns null;
	 */
	getValue: function()
	{
		AXES.SysTrace.error(TEXT.Dev("Attempt to call %s1 on a Select field ignored.", "getValue"));
		return null;
	},
	isModified: function() {
		return ((this.scrollAmount != undefined) || (this.data != this.initialData) || this.mdt)
	},
	/**
	 * Sets the terminal cursor position to the field's position.  This updates
	 * the internal representation of the screen and the cursor position that is
	 * sent to the host.  It does not affect the user interface in any way.  If you
	 * want to move the cursor in the user interface, use setFocus instead.
	 * @param {Integer} [offset=0] The select item within the field to place the cursor at.
	 */
	setCursor: function(offset) {
		offset = (isNaN(offset)) ? 0: parseInt(offset);
		var newPos = {
			col: this.col,
			row: this.row
		};
		if ((this.selectItems.length > 0) && (offset < this.selectItems.length)) 
		{
			newPos.col += this.selectItems[offset].col;
			newPos.row += this.selectItems[offset].row;
		}
		AXES.Terminal.cursor.set(newPos);
	},
	/**
	* Returns an array representing the fields footprint on the screen
	*/
	getTextArray: function()
	{
		var h = (typeof this.height == "number") ? this.height : 1;
		var w = (typeof this.width == "number") ? this.width : 0;
		var a = new Array(h);
		var s = this.data.split(/\n/);
		for (var i = 0; i < h; i++) 
		{
			a[i] = (s[i]) ? s[i].pad(w) : "".pad(w);
		}
		for (var i=0, sel; sel=this.selectItems[i]; i++) {
			a[sel.row] = a[sel.row].substring(0,sel.col) + sel.text.rpad(sel.width) + a[sel.row].substring(sel.col + sel.width);
		}
		return a;
	}
});

AXES.Terminal.Field.HtmlField = AXES.Terminal.Field.Base.extend(
/**
* @lends AXES.Terminal.Field.HtmlField.prototype
*/
{
	/** 
	 * Represents an HTML field in the 5250 screen
	 * @constructs
	 * @augments AXES.Terminal.Field.Base
	 * @param {Object} node A jQuery object containing the XML node from the server that defines the field.
	 */
	init: function(fieldNode)
	{
		this._super(fieldNode);
		if (fieldNode) {
			this.html = fieldNode.attr("html");
			this.scripts = [];	
			if (this.html == undefined) {
				this.html = AXES.Utils.xml2String(node);
				if (typeof this.html == "String") 
				{
					this.html = this.html.replace(/<\/?HTMLKW(.|\s)*?>/gm, "");
				// <![CDATA[ ]]> 
				
				} else {
					AXES.SysTrace.error(TEXT.Dev("Unable to process HTML embedded in DDS."));
				}
			}
			var self = this;
			this.html = this.html.replace(/<script(?:.|\s)*?>\s*(?:<\!\[CDATA\[)?((?:.|\s)*?)(?:\]\]>)?\s*<\/script>/gmi, function(match, p1, offset, s) {
				self.scripts.push(p1);
				return "";
			});		
		}
	},
	/**
	* Returns an empty array as an HTML field is not valid for a 5250 screen.
	*/
	getTextArray: function()
	{
		return [];
	}
});
